import pathlib
from femagtools.dxfsl.converter import convert

def test_convert():
    p = pathlib.Path(__file__).parent / 'data' / 'IPM-130-4.dxf'
    r = convert(str(p))

    assert r['num_poles'] == 4
    assert r['tot_num_slot'] == 12
    totnumsl = [l for l in r['fsl'] if l.startswith('m.tot_num_slot')]
    assert len(totnumsl) == 1
    assert totnumsl[0].split()[-1] == '12'



import femagtools.heat_source_network as hsn
import pathlib
import pytest


@pytest.fixture
def data_dir():
    return pathlib.Path(__file__).with_name('data')


def test_heat_source_network(data_dir):
    model = hsn.read(str(data_dir / 'temp_model.hsn'))
    assert [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 11, 13] == model.nodes
    names = model.get_node_names()
    assert ['StZa', 'outs', 'StJo', 'Slot',
            'Shaf', 'Iron', 'PMag', 'PMag',
            'W1  ', 'W2  ', 'W3  '] == names
    T = model.solve()
    # [130.22552926,  56.75737966, 116.12788806, 150.18881253,
    #   175.77455152, 175.77455152, 175.77526274, 175.77455152,
    #   126.04691063, 122.49609205, 125.27401647])



#!/usr/bin/env python
#
import unittest
import os
import femagtools.vbf


class VbfReaderTest(unittest.TestCase):
    def test_read_vbf(self):
        testPath = os.path.split(__file__)[0]
        if not testPath:
            testPath = '.'
            
        filename = "data/m270_35.vbf"
        vbf = femagtools.vbf.Reader('{0}/{1}'.format(testPath, filename))
        r = vbf.getLossValues()
        self.assertEqual(r['name'], 'M235_35')
        self.assertEqual(len(r['f']), 7)
        self.assertEqual(len(r['B']), 19)
        self.assertEqual(len(r['pfe']), 19)
        self.assertEqual(len(r['pfe'][0]), 7)
        self.assertAlmostEqual(r['f'][-1], 2000.0, 1)
        self.assertAlmostEqual(r['B'][-1], 1.926, 3)
        self.assertAlmostEqual(r['pfe'][3][-1], 46.633, 3)

if __name__ == '__main__':
    unittest.main()



import sys
import os
import pytest
import numpy as np
import femagtools.mcv
import pathlib

@pytest.fixture
def mcv_data_dir():
    return pathlib.Path(__file__).with_name('data')

def test_write_mcv_magcrv(tmpdir):
    filename = f"{tmpdir}/mcvData"

    B = [3e-02, 7e-02, 2e-01, 2.31]
    H = [1e+02, 2e+02, 5e+02, 3e+05]
    mcvData = dict(curve=[dict(bi=B,hi=H)])
    mcv = femagtools.mcv.MagnetizingCurve(mcvData)
    mcv.writefile(filename)
    mcvr = femagtools.mcv.Reader()
    ext = '.MC' if sys.platform == 'win32' else '.MCV'
    mcvr.readMcv(filename+ext)
    assert mcvr['version_mc_curve'] == 0
    mcvr['ctype'] == femagtools.mcv.MAGCRV
    assert mcvr['curve'][0]['bi'] == pytest.approx(B, 1e-3)
    assert mcvr['curve'][0]['hi'] == pytest.approx(H, 1e-3)
    assert len(mcvr['curve']) == 1

def test_write_mcv_orient_crv(tmpdir):
    filename = f"{tmpdir}/mcvData"

    B = [3e-02, 7e-02, 2e-01, 2.31]
    H = [1e+02, 2e+02, 5e+02, 3e+05]
    mcvData = dict(curve=[dict(bi=B,hi=H), dict(bi=B,hi=H)])
    mcv = femagtools.mcv.MagnetizingCurve(mcvData)
    mcv.writefile(filename)
    ext = '.MC' if sys.platform == 'win32' else '.MCV'
    mcvr = femagtools.mcv.Reader()
    mcvr.readMcv(filename+ext)
    assert mcvr['version_mc_curve'] == 1
    assert mcvr['ctype'] == femagtools.mcv.MAGCRV
    assert mcvr['curve'][0]['bi'] == pytest.approx(B, 1e-3)
    assert mcvr['curve'][0]['hi'] == pytest.approx(H, 1e-3)
    assert mcvr['curve'][1]['bi'] == pytest.approx(B, 1e-3)
    assert mcvr['curve'][1]['hi'] == pytest.approx(H, 1e-3)
    assert len(mcvr['curve']) == 2

def test_read_mcv_magcrv(mcv_data_dir):
    filename = "TKS_NO_20.MCV"
    reader = femagtools.mcv.Reader()
    reader.readMcv(mcv_data_dir / filename)
    r = reader.get_results()
    assert r['desc'] == u'PowerCore\xae NO 20 ;ThyssenKrupp Steel Eur'
    assert r['fillfac'] == pytest.approx(0.92, 1e-6)
    assert len(r['curve'][0]['bi']) == 24
    assert r['curve'][0]['bi'][0] == 0.0
    assert r['curve'][0]['bi'][-1] == pytest.approx(1.836, 1e-3)

def test_read_write_mcv_with_losses(mcv_data_dir):
    filename = "TKM270-50A-LOSS.MCV"
    reader = femagtools.mcv.Reader()
    reader.readMcv(mcv_data_dir / filename)
    r = reader.get_results()
    assert r['name'] == u'TKM270-50A-LOSS'
    assert len(r['curve'][0]['bi']) == 35
    assert r['curve'][0]['bi'][-1] == pytest.approx(2.638, 1e-3)
    assert set(r['losses'].keys()) == {'B', 'f', 'pfe', 'Bo', 'fo',
                                       'cw', 'cw_freq', 'b_coeff', 'ch', 'ch_freq'}
    assert len(r['losses']['pfe']) == 20
    assert len(r['losses']['pfe'][8]) == 19
    assert r['losses']['pfe'][8][18] == pytest.approx(3097.6, 0.1)

    # test mcv writer
    filename_out = "TKS_LOSS.MCV"
    writer = femagtools.mcv.Writer(r)
    writeMcvFile = mcv_data_dir / filename_out
    writer.writeMcv(writeMcvFile)
    # TEST
    reader2 = femagtools.mcv.Reader()
    reader2.readMcv(writeMcvFile)

    for attr in ['version_mc_curve', 'mc1_curves', 'mc1_title']:
        assert getattr(reader, attr) == getattr(reader2, attr)
    for attr in ['mc1_remz', 'mc1_bsat', 'mc1_bref', 'mc1_fillfac',
                 'fo', 'Bo', 'ch', 'ch_freq', 'cw',
                 'cw_freq', 'b_coeff', 'rho', 'fe_sat_mag']:
        assert getattr(reader, attr) == getattr(reader2, attr)
    for attr in ['hi', 'bi']:
        assert reader.curve[0][attr] == reader2.curve[0][attr]
    np.testing.assert_almost_equal(reader.losses['f'],
                                   reader2.losses['f'], 5)
    np.testing.assert_almost_equal(reader.losses['B'],
                                   reader2.losses['B'], 5)

    writer.bertotti = {'ch': 0.02196930496538005,
                      'cw': 9.739048407022554e-05,
                       'ce': 0.00047834510925596726,
                       'ch_freq': 2, 'b_coeff':0}
    writer.writeMcv(mcv_data_dir / "TKS_LOSS2.MCV", feloss='bertotti')


def test_extrapol():
    mcv = {'name':"M222",
           'curve':[{
            'bi':[0.0, 0.09, 0.179, 0.267, 0.358,
                0.45, 0.543, 0.6334, 0.727,
                0.819, 0.9142, 1.0142, 1.102,
                1.196, 1.314, 1.3845, 1.433,
                1.576, 1.677, 1.745, 1.787,
                1.81, 1.825, 1.836],

            'hi':[0.0, 22.16, 31.07, 37.25, 43.174,
                49.54, 56.96, 66.11, 78.291,
                95, 120.64, 164.6, 259.36,
                565.86, 1650.26, 3631.12, 5000, 10000,
                15000, 20000, 25000, 30000, 35000, 40000]}]}
    c = femagtools.mcv.extrapol(mcv['curve'][0])
    assert len(c['bi']) > len(mcv['curve'][0]['bi'])
    assert len(c['hi']) > len(mcv['curve'][0]['hi'])
    assert np.all(np.diff(c['bi']) > 0)
    assert np.all(np.diff(c['hi']) > 0)



import numpy as np 
from .utils import fft
import logging 

logger = logging.getLogger('femagtools.semi_fea')

def shift_array(v, idx): 
    '''shift array by index'''
    return v[idx::] + v[0:idx]

def fft_filter(result_fft, perc=0.01):
    '''filter FFT result with amplitude'''
    result = {"order": [], "y":[]}
    base_amp = result_fft['a']
    for i, j in enumerate(result_fft['nue']):
        if j >= perc*base_amp: 
            result['order'].append(i)
            result['y'].append(j)
    return result

def fast_skew_cogg(result, skew_setup): 
    '''Calculate cogging torque/Back-EMF with step skewing based on unskewed result
    Arguments: 
    result: BCH objects 
    skew_setup(dict): {"skew_angle": 10, "nu_skew_steps": 2}
    '''
    skew_angle = skew_setup['skew_angle']
    num_skew_steps =skew_setup['num_skew_steps']
    skew_angle_intern = 0.0
    skew_angle_array = []
    T_slice = []
    bemf = {"1": [], "2": [], "3": []}
    bemf_slice = {"1": [], "2": [], "3": []}
    bemf_skew = {"1": [], "2": [], "3": []}
    bemf_skew_fft = {"1": [], "2": [], "3": []}

    keyset = ('1', '2', '3')

    # check if skew steps equals 2
    if num_skew_steps == 2: 
        skew_angle_intern = skew_angle
        skew_angle_array = [-skew_angle_intern/2, skew_angle_intern/2]
    else: 
        skew_angle_intern = skew_angle/num_skew_steps*(num_skew_steps-1)/2
        skew_angle_array = np.linspace(-skew_angle_intern, skew_angle_intern, 
                                       num_skew_steps, endpoint=True).tolist()

    angle = result.torque[-1]['angle']
    T = result.torque[-1]['torque'][0:-1]
    # get back-emf from BCH
    for i in keyset: 
        bemf[i] = result.flux[i][-1]['voltage_dpsi'][0:-1]

    angl_resl = angle[1]
    tmp = np.unique(np.abs(skew_angle_array))
    skew_angl_resl = 0.0
    if np.amin(tmp) == 0.0: 
        skew_angl_resl = tmp[1]
    else: 
        skew_angl_resl = tmp[0]
    
    divider = skew_angl_resl/angl_resl 
    if divider - np.floor(divider) > 1e-15: 
        # TODO: Interpolation if angle resolution doesn't match
        logger.warning("Wrong Mesh Size in the airgap mesh")
    else: 
        logger.info(f"number of element shifted {divider}")

    for i in skew_angle_array: 
        idx = int(i/angl_resl)
        if i != 0: 
            T_slice.append(shift_array(T, idx))
            for j in keyset:
                bemf_slice[j].append(shift_array(bemf[j], idx))
        else: 
            # do nothing
            T_slice.append(T)
            for j in keyset:
                bemf_slice[j].append(bemf[j])

    # average torque
    T_sum = 0
    for i in T_slice: 
        T_sum += np.array(i)
    T_skew = (T_sum/num_skew_steps).tolist()
    T_skew += [T_skew[0]]
    T_fft = fft_filter(fft(angle, T_skew, pmod=2))

    # average back-emf
    for j in keyset:
        flx_skew = 0
        for k in bemf_slice[j]:
            flx_skew+=np.array(k)
        bemf_skew[j] = (flx_skew/num_skew_steps).tolist()
        bemf_skew[j] += [bemf_skew[j][0]]
        bemf_skew_fft[j] = fft_filter(fft(angle, bemf_skew[j], pmod=2))

    for i in range(len(T_slice)):
        T_slice[i] = (np.array(T_slice[i])/num_skew_steps).tolist()
        T_slice[i]+=[T_slice[i][0]]
    
    return {"angle": angle,
            "cogging_torque": T_skew, 
            "cogging_torque_fft": T_fft, 
            "BEMF": bemf_skew, 
            "BEMF_fft": bemf_skew_fft, 
            "cogging_torque_slice": T_slice}
    


# -*- coding: utf-8 -*-
"""
  femagtools.dxfsl.symmetry
  ~~~~~~~~~~~~~~~~~~~~~~~~~

  Authors: Ronald Tanner, Beat Holm
"""
from __future__ import print_function

import numpy as np
import logging
import sys
from femagtools.dxfsl.shape import Element, Line
from femagtools.dxfsl.area import Area
import femagtools.dxfsl.area as AREA
from femagtools.dxfsl.functions import alpha_angle, positive_angle, is_same_angle
from femagtools.dxfsl.functions import min_angle, max_angle, gcd, point
from femagtools.dxfsl.functions import less_equal, less, points_are_close
from femagtools.dxfsl.functions import line_m, line_n, mirror_point
from femagtools.dxfsl.functions import part_of_circle, round_point

logger = logging.getLogger('femagtools.symmetry')


#############################
#         symmetry          #
#############################


class Symmetry(object):
    def __init__(self,
                 geom=None,
                 startangle=None,
                 endangle=None,
                 rtol=1e-03,
                 atol=1e-02):
        assert(geom is not None)
        assert(startangle is not None)
        assert(endangle is not None)
        
        self.geom = geom
        self.startangle = startangle
        self.endangle = endangle
        self.geom_part = part_of_circle(self.startangle, self.endangle, 1)
        self.delta_check_count = 0
        self.delta_angle_corr = None
        self.rtol = rtol
        self.atol = atol
        self.full = False
        self.ag_radius = 0.0
        self.height = geom.max_radius - geom.min_radius
        if np.isclose(self.startangle, self.endangle):
            self.alpha = 2.0*np.pi
            self.full = True
        else:
            self.alpha = alpha_angle(self.startangle, self.endangle)
            self.full = False
        if geom.is_inner:
            self.ag_radius = geom.max_radius
        else:
            self.ag_radius = geom.min_radius
        logger.debug("Symmetry(alpha=%s, rtol=%s, atol=%s)", self.alpha, rtol, atol)

    def __str__(self):
        return "rtol: {}\n".format(self.rtol) + \
               "atol: {}\n".format(self.atol)

    def equal_area(self,
                   d1, h1, a1,
                   d2, h2, a2,
                   rtol=1e-03, atol=1e-03):
        equal_d = np.isclose(d1, d2, rtol=rtol, atol=atol)  # distance form c
        equal_h = np.isclose(h1, h2, rtol=rtol, atol=atol)  # height
        equal_a = np.isclose(a1, a2, rtol=1e-4, atol=1e-2)  # angle from c
        if not equal_d:
            logger.debug("equal_area: dist NOT close (%s/%s)", d1, d2)
            if equal_h and equal_a:  # but height and angle
                if np.isclose(d1, d2, rtol=1e-2, atol=1.e-1):
                    logger.debug(" -- but with more tolerance")
                    return True
            return False
        if not equal_h:
            logger.debug("equal_area: height NOT close (%s/%s)", h1, h2)
            if equal_d and equal_a:  # but distance and angle
                if np.isclose(h1, h2, rtol=1e-2, atol=1e-1) :
                    logger.debug(" -- but with more tolerance")
                    return True
            return False
        if not np.isclose(a1, a2, rtol=1e-4, atol=1e-2):
            logger.debug("equal_area: alpha NOT close (%s/%s)", a1, a2)
            return False
        else:
            if a1 > a2:
                f = a2 / a1
            else:
                f = a1 / a2
            if f < 0.9:
                return False
        return True

    def calc_mid_angle(self, a):
        return positive_angle(alpha_angle(self.startangle,
                                          a.get_mid_angle(self.geom.center)))

    def area_list_entry(self, a):
        a.set_symmetry_parameter(self.geom.center)
        return (round(a.get_alpha(self.geom.center), 3),
                round(a.min_dist, 1),
                round(a.height, 1),
                self.calc_mid_angle(a),
                a)

    def build_area_list(self, types=()):
        arealist = self.geom.list_of_areas()
        if types:
            arealist = [a for a in arealist if a.type in types]

        areas = []
        for a in arealist:
            areas.append(self.area_list_entry(a))
        areas.sort(reverse=True)
        return areas

    def build_results(self, areas):
        logger.debug("begin of build_results with %s areas", len(areas))
        [logger.debug("#%s: alpha=%s, min=%s, max=%s",
                      a.get_id(), alpha, a.min_angle, a.max_angle) for
         alpha, dist, h, mid, a in areas]

        a0_alpha, a0_min_dist, a0_height, a0_mid_angle, a0 = areas[0]
        equal_areas = [(a0_mid_angle, a0)]
        check_rslt = []
        for a1_alpha, a1_min_dist, a1_height, a1_mid_angle, a1 in areas[1:]:
            if (self.equal_area(a0_min_dist, a0_height, a0_alpha,
                                a1_min_dist, a1_height, a1_alpha,
                                rtol=0.001, atol=0.05)):
                a0_min_dist = (a0_min_dist + a1_min_dist) / 2
                a0_height = (a0_height + a1_height) / 2
                a0_alpha = (a0_alpha + a1_alpha) / 2
                equal_areas.append((a1_mid_angle, a1))
            else:
                # alpha Wechsel
                id_list = []
                for i in range(len(equal_areas)):
                    mid_angle0, area0 = equal_areas[i]
                    if area0.get_id() in id_list:
                        continue
                    equal_areas_check = [(mid_angle0, area0)]
                    for mid_angle1, area1 in equal_areas[i+1:]:
                        if area1.get_id() in id_list:
                            continue
                        if area0.is_symmetry_equal(area1):
                            equal_areas_check.append((mid_angle1, area1))
                            id_list.append(area1.get_id())

                    rslt = self.check_delta(equal_areas_check)
                    areasize = a0.area_size()
                    rslt['area'] = a0
                    rslt['areasize'] = areasize
                    check_rslt.append((areasize, rslt))

                equal_areas = [(a1_mid_angle, a1)]
                a0_min_dist = a1_min_dist
                a0_height = a1_height
                a0_alpha = a1_alpha
                a0 = a1

        rslt = self.check_delta(equal_areas)
        areasize = a0.area_size()
        rslt['area'] = a0
        rslt['areasize'] = areasize
        check_rslt.append((areasize, rslt))
        logger.debug("end of build_results")
        return check_rslt

    def get_winding_symmetry(self, inside=False):
        if inside:
            areas = [self.area_list_entry(a) for a in self.geom.list_of_areas()
                     if not a.close_to_ag]
            areas.sort(reverse=True)
        else:
            areas = self.build_area_list((AREA.TYPE_WINDINGS,))

        logger.debug("begin of Symmetry::get_winding_symmetry: %s areas available", len(areas))
        if not areas:
            logger.debug("end of Symmetry::get_winding_symmetry: no areas")
            return 0

        check_rslt = self.build_results(areas)
        logger.debug("%s results available", len(check_rslt))
        [logger.debug("Result: %s", rslt) for rslt in check_rslt]

        parts, start_delta = self.get_symmetry_parts(check_rslt)
        if parts <= 1:
            return 0
        self.create_cut_lines(parts, start_delta)

        sym = self.geom_part * parts
        delta = 2*np.pi/sym
        self.set_symmetry_parameters(self.startangle, parts, delta)

        logger.debug("end of Symmetry::get_winding_symmetry: parts=%s", parts)
        return parts

    def get_magnet_symmetry(self):
        areas = self.build_area_list((AREA.TYPE_MAGNET_AIRGAP, AREA.TYPE_MAGNET_RECT,))
        air = self.build_area_list((AREA.TYPE_AIR,))
        mag_list = [a for a in self.geom.list_of_areas() if a.is_magnet()]
        air_list = [a for a in self.geom.list_of_areas() if a.is_air()]
        sz_list = [a.area_size() for a in self.geom.list_of_areas()]
        max_sz = max(sz_list)
        for a in air_list:
            if a.area_size() < max_sz * 0.005:
                continue
            for m in mag_list:
                if a.is_touching(m):
                    areas.append(self.area_list_entry(a))
                    break

        logger.debug("begin of Symmetry::get_magnet_symmetry: %s areas available", len(areas))
        if not areas:
            logger.debug("end of Symmetry::get_magnet_symmetry: no areas")
            return 0

        check_rslt = self.build_results(areas)
        logger.debug("%s results available", len(check_rslt))
        [logger.debug("Result: %s", rslt) for rslt in check_rslt]
        for sz, rslt in check_rslt:
            if not rslt.get('startdelta', 0.0) == 0.0:
                return 0  # not proper
            if rslt.get('halfslice', None):
                return 0  # not proper

        parts, start_delta = self.get_symmetry_parts(check_rslt)
        if parts <= 1:
            return 0
        self.create_cut_lines(parts, start_delta)

        sym = self.geom_part * parts
        delta = 2*np.pi/sym
        self.set_symmetry_parameters(self.startangle, parts, delta)

        logger.debug("end of Symmetry::get_magnet_symmetry: parts=%s", parts)
        return parts

    def find_symmetry(self):
        areas = self.build_area_list()

        logger.debug("begin of Symmetry::find_symmetry: %s areas available", len(areas))
        if not areas:
            logger.debug("end of Symmetry::find_symmetry: no areas")
            return 0

        check_rslt = self.build_results(areas)
        logger.debug("%s results available", len(check_rslt))

        parts, start_delta = self.get_symmetry_parts(check_rslt)
        if parts < 2:
            logger.debug("end of Symmetry::find_symmetry: no symmetry")
            return parts

        if self.delta_angle_corr is not None and self.delta_angle_corr != 0.0:
            self.startangle = self.startangle - self.delta_angle_corr
            self.endangle = self.endangle - self.delta_angle_corr

        self.create_cut_lines(parts, start_delta)

        logger.debug("end of Symmetry::find_symmetry: -> %s", parts)
        return parts

    def check_delta(self, area_list):
        logger.debug("begin of check_delta: %s equal areas", len(area_list))
        result = {'areas': len(area_list),
                  'startdelta': 0.0,
                  'slices': None}
        result['area_id_list'] = [a.get_id() for m, a in area_list]
        if not area_list:
            logger.debug("end of check_delta: no areas")
            return result

        rtol = 1e-3
        atol = 1e-2

        logger.debug("Geometry: Alpha=%s,  Center=%s", self.alpha, self.geom.center)
        mid_angle, a = area_list[0]
        result['height'] = a.height
        result['min_dist'] = a.min_dist
        result['max_dist'] = a.max_dist
        result['alpha'] = a.get_alpha(self.geom.center)
        if self.geom.is_inner:
            result['airgap'] = np.isclose(a.max_dist, self.ag_radius)
        else:
            result['airgap'] = np.isclose(a.min_dist, self.ag_radius)
            upper_quarter = self.geom.min_radius + self.height * 0.75
            upper_half = self.geom.min_radius + self.height * 0.5
            if a.max_dist > upper_quarter:
                if a.min_dist > upper_half:
                    result['upper_quarter'] = True
        if len(area_list) == 1:  # one area
            rslt = self.check_one_area(mid_angle, a, result, rtol=rtol, atol=atol)
            logger.debug("end of check_delta")
            return rslt

        self.delta_check_count += 1
        area_list.sort()

        mid_angle, a = area_list[0]
        delta = positive_angle(mid_angle * 2)
        delta_total = mid_angle
        logger.debug("first delta=%s,  total=%s", delta, delta_total)
        delta_list = [delta]
        mid_delta_list = [(mid_angle, delta)]
        logger.debug("First mid = %s, delta = %s", mid_angle, delta)
        logger.debug("%s:  d=%s,  h=%s,  a=%s, mid=%s, delta=%s",
                     a.identifier(),
                     a.min_dist,
                     a.height,
                     a.get_alpha(self.geom.center),
                     mid_angle,
                     delta)

        geom_alpha = alpha_angle(self.startangle, self.endangle)
        geom_alpha = positive_angle(geom_alpha)

        start_angle = mid_angle
        for mid_angle, a in area_list[1:]:
            delta_angle = alpha_angle(start_angle, mid_angle)
            delta = positive_angle(delta_angle)
            delta_total += delta_angle
            logger.debug("next delta=%s,  total=%s", delta, delta_total)

            delta_list.append(delta)
            mid_delta_list.append((mid_angle, delta))

            logger.debug("%s:  d=%s,  h=%s,  a=%s, mid=%s, delta=%s",
                         a.identifier(),
                         a.min_dist,
                         a.height,
                         a.get_alpha(self.geom.center),
                         mid_angle,
                         delta)
            start_angle = mid_angle
            
        delta_angle = alpha_angle(start_angle, geom_alpha)
        if np.isclose(delta_angle, np.pi*2, rtol=1e-4, atol=1e-4):
            logger.debug("Last Area is in the middle of endangle")
            delta_angle = 0.0
        delta = positive_angle(delta_angle * 2)
        delta_total += delta_angle
        delta_list.append(delta)
        mid_delta_list.append((0.0, delta))

        logger.debug("final delta=%s,  total=%s", delta, delta_total)
        logger.debug("Delta List: %s", delta_list)
        logger.debug("Mid Delta List")
        [logger.debug("-- Mid angle: %s,   Delta: %s", a, d) for a, d in mid_delta_list]

        if not np.isclose(geom_alpha, delta_total, rtol=rtol, atol=atol):
            logger.debug("-- deltas: %s", delta_list)
            logger.debug("end of check_delta: BAD DELTA %s, (expected %s)",
                         delta_angle, geom_alpha)
            result['slices'] = 0
            logger.debug("end of check_delta: very bad")
            return result  # very bad

        deltas = self.create_deltas(delta_list, rtol=rtol, atol=atol)

        logger.debug("Start with looking for symmetry")

        logger.debug(">> %s Deltas <<", len(deltas))
        [logger.debug(" -- n=%s,  delta=%s", n, d) for n, d in deltas]

        if len(deltas) == 2:
            n1, d1 = deltas[0]
            n2, d2 = deltas[1]
            logger.debug("delta 1: n=%s,  delta=%s", n1, d1)
            logger.debug("delta 2: n=%s,  delta=%s", n2, d2)

            if n2 == 2 and n1 > 2 and \
               np.isclose(d1, d2 / 2.0, rtol=rtol, atol=atol):
                if np.isclose(d2, delta_list[0], rtol=rtol, atol=atol) and \
                   np.isclose(d2, delta_list[-1], rtol=rtol, atol=atol):
                    slices = n1 + n2
                    if slices > 4:
                        result['slices'] = slices
                    else:
                        result['slices'] = 1
                    result['slices_half'] = slices
                    result['halfslice'] = 1
                    result['startdelta'] = d1 / 2
                    logger.debug("#3: end of check_delta: SYMMETRY FOUND [%s] halfslice",
                                 result['slices'])
                    return result

            elif n1 == 2 and n2 == 1:
                assert(len(area_list) == 2)
                semi_alpha = result['alpha'] / 2
                a0_mid_angle, a0 = area_list[0]
                a1_mid_angle, a1 = area_list[1]
                a0_start = np.isclose(a0_mid_angle - semi_alpha,
                                      0.0,
                                      rtol=rtol, atol=atol)
                a1_end = np.isclose(a1_mid_angle + semi_alpha,
                                    self.alpha,
                                    rtol=rtol, atol=atol)
                if a0_start and a1_end and d1 < d2:
                    result['slices'] = 0
                    result['halfslice'] = 2
                    logger.debug("#4: end of check_delta: half slices")
                    return result

                if not a0_start and not a1_end and d1 > d2:
                    parts_in_circ = part_of_circle(0.0, d2, 1)
                    parts_in_geom = float(round(parts_in_circ / self.geom_part, 2))
                    if parts_in_geom.is_integer():
                        parts_in_geom = int(parts_in_geom)
                    else:
                        parts_in_geom = 0
                    result['slices'] = parts_in_geom
                    result['slices_half'] = parts_in_geom
                    result['halfslice'] = 1
                    logger.debug("#5: end of check_delta: SYMMETRY FOUND [%s] halfslice",
                                 result['slices'])
                    return result

            elif abs(n1 - n2) == 1:
                if d1 < d2:
                    if self.check_pairs_in_delta_list(delta_list,
                                                      deltas,
                                                      rtol=rtol, atol=atol):
                        result['slices'] = int(len(area_list) / 2)
                        delta_angle_corr = (d1 + d2) / 2
                        result['delta_corr'] = delta_angle_corr
                        logger.debug("Startangle correction by %s", delta_angle_corr)
                        logger.debug("#6: end of check_delta: SYMMETRY FOUND [%s]",
                                     result['slices'])
                        return result

        if len(deltas) > 1:
            n1, d1 = deltas[0]
            n2, d2 = deltas[1]
            if n1 + n2 + 1 == len(area_list) and abs(n1 - n2) == 1:
                dlist = self.check_pairs_of_areas(mid_delta_list,
                                                  min(d1, d2),
                                                  geom_alpha,
                                                  rtol=rtol, atol=atol)
                if dlist:
                    delta_list = dlist
                    deltas = self.create_deltas(delta_list, rtol=rtol, atol=atol)

        if False:
            parts_in_circ = part_of_circle(0.0, d1, 1)
            parts_in_geom = float(round(parts_in_circ / self.geom_part, 2))
            if parts_in_geom.is_integer():
                parts_in_geom = int(parts_in_geom)
            else:
                parts_in_geom = 0

            if parts_in_geom / n1 > 0.75 and parts_in_circ > 15:
                result['slices'] = parts_in_geom
                logger.debug("#7: end of check_delta: SYMMETRY FOUND [%s]",
                             result['slices'])
                return result

        missing_middle = []
        if len(deltas) == 2:
            logger.debug("looking for holes in delta list")

            delta_n, delta_value = deltas[0]
            logger.debug("First n,v == (%s, %s)", delta_n, delta_value)
            for n, v in deltas[1:]:
                logger.debug("Next n,v == (%s, %s)", n, v)
                if n < delta_n / 4 and \
                   np.isclose(delta_value, v / 2.0, rtol=1e-04, atol=1e-03):
                    logger.debug("Hole found")
                    inx = [i for i, x in enumerate(delta_list)
                           if np.isclose(x, v, rtol=rtol, atol=atol)]
                    if len(inx) != n:
                        logger.debug("Hole missmatch: %s <> %s", len(inx), n)
                        result['slices'] = 0
                        return result

                    dlist = []
                    x = 0

                    for i in inx:
                        for n in range(x, i):
                            logger.debug("set value of index %s", n)
                            dlist.append(delta_list[n])
                        m1, a = area_list[i-1]
                        if i < len(area_list):
                            m2, a = area_list[i]
                        else:
                            m2, a = area_list[0]
                        mid = (m1 + m2) / 2
                        logger.debug("Missing mid is %s", mid)
                        missing_middle.append((mid, i))
                        x = i+1
                        logger.debug("set value in hole")
                        dlist.append(delta_value)
                        dlist.append(delta_value)
                    for n in range(x, len(delta_list)):
                        logger.debug("set value of index %s", n)
                        dlist.append(delta_list[n])
                    logger.debug("New List: %s", dlist)
                    delta_list = dlist

        result['middlelist'] = [m for m, a in area_list]
        result['missing_middles'] = missing_middle

        if not np.isclose(delta_list[0], delta_list[-1], rtol=rtol, atol=atol):
            logger.debug("First and Last delta not equal")
            if self.full:
                d0 = (delta_list[0] + delta_list[-1]) / 2
                n1, d1 = deltas[0]
                if np.isclose(d0, d1, rtol=rtol, atol=atol):
                    delta_angle_corr = (d0 - delta_list[0]) / 2
                    result['delta_corr'] = delta_angle_corr
                    logger.debug("Startangle correction by %s", delta_angle_corr)
                    delta_list[0] = d0
                    delta_list[-1] = d0
            else:
                parts = self.check_first_last_difference(delta_list, deltas)
                if parts > 0:
                    result['slices'] = parts
                    logger.debug("#8: end of check_delta: SYMMETRY FOUND [%s]",
                                 result['slices'])
                    return result

        logger.debug("Final Delta List: %s", delta_list)
        d1 = delta_list[0]
        d1_count = 1
        inx_list = [0]
        for x in range(1, len(delta_list)):
            if np.isclose(d1, delta_list[x], rtol=rtol, atol=atol):
                inx_list.append(x)
                d1_count += 1

        if d1_count == len(delta_list):
            result['slices'] = d1_count -1
            logger.debug("#9: end of check_delta: SYMMETRY FOUND [%s]",
                         result['slices'])
            return result  # very simple
        if len(delta_list) < 3:
            logger.debug("end of check_delta: One delta only ?!")
            result['slices'] = 0
            return result

        logger.debug("index of delta %s: %s", d1, inx_list)
        if len(inx_list) < 2:
            logger.debug("end of check_delta: NO SYMMETRY")
            result['slices'] = 0
            return result

        x1 = inx_list[0]
        x2 = inx_list[1]
        step = x2 - x1
        x1 = x2
        for x2 in inx_list[2:]:
            if not (x2 - x1 == step):
                logger.debug("end of check_delta: NO SYMMETRY")
                result['slices'] = 0
                return result
            x1 = x2

        logger.debug("length of delta %s: %s",
                     len(inx_list),
                     inx_list)
        result['slices'] = len(inx_list) -1
        logger.debug("#10: end of check_delta: SYMMETRY FOUND [%s]", result['slices'])
        return result

    def create_deltas(self, delta_list, rtol=1e-3, atol=1e-2):
        delta_list_sorted = [d for d in delta_list]
        delta_list_sorted.sort()
        delta = delta_list_sorted[0]
        delta_n = 1
        deltas = []
        for i in range(1,len(delta_list_sorted)):
            if np.isclose(delta_list_sorted[i], delta, rtol=rtol, atol=atol):
                delta = (delta + delta_list_sorted[i]) / 2
                delta_n += 1
            else:
                deltas.append((delta_n, delta))
                delta = delta_list_sorted[i]
                delta_n = 1
        deltas.append((delta_n, delta))
        deltas.sort(reverse=True)
        return deltas

    def check_one_area(self, mid_angle, a, result, rtol=1e-3, atol=1e-2):
        logger.debug("begin of check_one_area")

        alpha = a.get_alpha(self.geom.center)
        logger.debug("Single %s:  d=%s,  h=%s,  a=%s, mid=%s",
                     a.identifier(),
                     a.min_dist,
                     a.height,
                     a.get_alpha(self.geom.center),
                     mid_angle)
        if np.isclose(alpha, self.alpha, rtol=rtol, atol=atol):
            logger.debug("end of check_one_area: area %s from start to end",
                         a.identifier())
            result['slices'] = None
            return result  # ok

        if self.full:
            result['slices'] = 1
            logger.debug("end of check_one_area: full with 1 slice")
            return result

        delta_angle = alpha_angle(self.startangle, mid_angle)
        delta1 = positive_angle(delta_angle)
        delta_angle = alpha_angle(mid_angle, self.endangle)
        delta2 = positive_angle(delta_angle)
        if np.isclose(delta1, delta2, rtol=rtol, atol=atol):
            result['slices'] = 1
            result['slices_half'] = 2
            result['halfslice'] = 1
            logger.debug("end of check_delta: One Area in the middle")
        else:
            result['middlelist'] = [mid_angle]
            result['slices'] = 0
            logger.debug("end of check_one_area: Area somewhere")
        return result

    def check_pairs_in_delta_list(self,
                                  delta_list,
                                  deltas,
                                  rtol=1e-3, atol=1e-2):
        if len(deltas) < 2:
            return False
        if len(delta_list) < 5:
            return False
        n1, d1 = deltas[0]
        n2, d2 = deltas[1]
        if not abs(n1 - n2) == 1:
            return False
        # check without first/last
        delta0 = delta_list[1]
        delta1 = delta_list[2]
        for delta2 in delta_list[3:-1]:
            if not np.isclose(delta0, delta2, rtol=rtol, atol=atol):
                return False
            delta0 = delta1
            delta1 = delta2
        logger.debug("** Pairs available **")
        return True

    def check_pairs_of_areas(self,
                             mid_delta_list,
                             delta,
                             geom_alpha,
                             rtol=1e-3, atol=1e-2):
        logger.debug("begin of check_pairs_of_areas")
        if len(mid_delta_list) < 2:
            return None

        logger.debug("Mid-Delta-List")
        [logger.debug(" -- mid=%s,  delta=%s",m, d) for m, d in mid_delta_list]

        # check
        mid_list = []
        m0, d0 = mid_delta_list[0]
        m1, d1 = mid_delta_list[1]
        if np.isclose(delta, d1, rtol=rtol, atol=atol):
            mx = (m0 + m1) / 2
            mid_list.append(mx)

        m0, d0 = mid_delta_list[2]
        dx = d1
        for m2, d2 in mid_delta_list[2:-1]:
            #logger.debug("compare %s and %s", d0, d2)
            if not np.isclose(d0, d2, rtol=rtol, atol=atol):
                logger.debug("end of check_pairs_of_areas: bad pairs")
                return None
            if np.isclose(delta, d2, rtol=rtol, atol=atol):
                mx = (m1 + m2) / 2
                mid_list.append(mx)
            d0 = d1
            m1 = m2
            d1 = d2

        if not mid_list:
            return None

        logger.debug("New Mids: %s", mid_list)
        delta = positive_angle(mid_list[0] * 2)
        delta_list = [delta]
        m0 = mid_list[0]
        for m1 in mid_list[1:]:
            delta = positive_angle(alpha_angle(m0, m1))
            delta_list.append(delta)
            m0 = m1

        delta_angle = alpha_angle(m1, geom_alpha)
        if np.isclose(delta_angle, np.pi*2, rtol=1e-4, atol=1e-4):
            logger.debug("Last Area is in the middle of endangle")
            delta_angle = 0.0
        delta = positive_angle(delta_angle * 2)
        delta_list.append(delta)
        logger.debug("New delta-list: %s", delta_list)
        logger.debug("end of check_pairs_of_areas")
        return delta_list

    def check_first_last_difference(self, delta_list, deltas):
        logger.debug("begin check_first_last_difference")
        if np.isclose(delta_list[0], delta_list[-1],
                      rtol=self.rtol, atol=self.atol):
            logger.debug("end check_first_last_difference: first/last equal")
            return 0
        if len(deltas) != 3:
            logger.debug("end check_first_last_difference: not 3 deltas")
            return 0
        logger.debug(">> 3 Deltas <<")
        n1, d1 = deltas[0]
        n2, d2 = deltas[1]
        n3, d3 = deltas[2]
        if not (n2 == 1 and n3 == 1):
            logger.debug("end check_first_last_difference: first/last diff")
            return 0
        dx = (d2 + d3) / 2
        if not np.isclose(dx, d1, rtol=self.rtol, atol=self.atol):
            logger.debug("end check_first_last_difference: bad deltas")
            return 0
        dx = (delta_list[0] + delta_list[-1]) / 2
        if not np.isclose(dx, d1, rtol=self.rtol, atol=self.atol):
            logger.debug("end check_first_last_difference: bad deltas")
            return 0

        logger.debug("end check_first_last_difference => %s", n1 + 1)
        return n1 + 1

    def concatenate_results(self, check_rslt):
        if len(check_rslt) < 2:
            return check_rslt
        #   ----------
        def get_slices(rslt):
            if rslt['slices'] is None:
                return 0
            if rslt.get('startdelta', 0.0) != 0.0:
                return 0
            if rslt.get('halfslice', None) is not None:
                return 0
            return rslt['slices']
        #   ------
        new_rslt = []
        size1, n1, rslt1 = check_rslt[0]
        slices1 = get_slices(rslt1)
        for size2, n2, rslt2 in check_rslt[1:]:
            slices2 = get_slices(rslt2)
            if slices1 > 0 and \
               slices1 == slices2 and \
               np.isclose(size1, size2, rtol=1e-01, atol=1e-01) and \
               (rslt1.get('delta_corr', 0.0) == 0.0 or \
                rslt2.get('delta_corr', 0.0) == 0.0):
                # concatenate
                rslt1['slices'] = None
                rslt2['delta_corr'] = 0.0
                rslt2['areas'] = rslt2['areas'] + rslt1['areas']
                logger.debug("concatenate results(size %s)", size1)
            else:
                new_rslt.append((size1, n1, rslt1))
            slices1, size1, n1, rslt1 = (slices2, size2, n2, rslt2)

        new_rslt.append((size1, n1, rslt1))
        return new_rslt

    def get_symmetry_parts(self, check_rslt):
        max_size = 0
        max_areas = 0
        max_slices = 0
        parts_possible = None
        self.delta_angle_corr = None
        unsure_sym = False

        check_rslt = [(size, n, rslt) for n, (size, rslt) in enumerate(check_rslt)
                      if rslt['slices'] is not None]
        check_rslt.sort(reverse=True)
        check_rslt = self.concatenate_results(check_rslt)
        [logger.debug("Result #%s: %s, %s", n, size, rslt) for size, n, rslt in check_rslt]

        rtol = 1e-3
        atol = 1e-2

        missing_middles = []
        halfslice = []
        start_delta = None
        start_delta_corr = 0.0

        with_angle_corr = 0
        without_angle_corr = 0
        maybe_angle_korr = 0

        for size, n, rslt in check_rslt:
            areas = rslt['areas']
            slices = rslt['slices']
            if slices is None:
                continue

            size = rslt['areasize']
            angle_corr = rslt.get('delta_corr', 0.0)

            if rslt.get('halfslice', 0) == 1:
                halfslice.append(rslt)

            if slices > 0:
                max_size = max(max_size, size)
                max_areas = max(max_areas, areas)
                max_slices = max(max_slices, slices)
                missing_middles += rslt.get('missing_middles', [])
                area = rslt['area']
                if not (np.isclose(area.min_dist,
                                   self.geom.min_radius,
                                   rtol=1e-4, atol=1e-3) and \
                        np.isclose(area.max_dist,
                                   self.geom.max_radius,
                                   rtol=1e-4, atol=1e-3)):
                    if rslt.get('delta_corr', 0.0) == 0.0:
                        without_angle_corr += areas * size
                    else:
                        with_angle_corr += areas * size
                else:
                    maybe_angle_korr += areas * size

        logger.debug("max size: %s,  max areas: %s", max_size, max_areas)
        logger.debug("Angle-Corrections: %s Yes,  %s No, %s Maybe",
                     with_angle_corr,
                     without_angle_corr,
                     maybe_angle_korr)

        if np.isclose(with_angle_corr, without_angle_corr):
            with_angle_corr = (maybe_angle_korr > 0)
        else:
            with_angle_corr = (with_angle_corr > without_angle_corr)

        #   -------------------------
        def get_halfslice_counterpart(rslt):
            if rslt.get('halfslice', 0) != 2:
                return None
            for half in halfslice:
                alpha1 = half['alpha']
                height1 = half['height']
                alpha2 = rslt['alpha'] * 2.0
                height2 = rslt['height']
                logger.debug("-- height: %s / %s", height1, height2)
                logger.debug("-- alpha: %s / %s", alpha1, alpha2)
                if np.isclose(height1, height2, rtol=rtol, atol=atol) and \
                   np.isclose(alpha1, alpha2, rtol=rtol, atol=atol):
                    return half
            return None
        #   -----------
        if halfslice:
            logger.debug("%s halfslice [1] found", len(halfslice))

            for size, n, rslt in check_rslt:
                half = get_halfslice_counterpart(rslt)
                if half:
                    logger.debug("Halfslice counterpart found")
                    alpha1 = half['alpha']
                    height1 = half['height']
                    alpha2 = rslt['alpha'] * 2.0
                    height2 = rslt['height']
                    logger.debug("-- height: %s / %s", height1, height2)
                    logger.debug("-- alpha: %s / %s", alpha1, alpha2)
                    if np.isclose(height1, height2, rtol=rtol, atol=atol) and \
                       np.isclose(alpha1, alpha2, rtol=rtol, atol=atol):
                        logger.debug("halfslice result: %s", half)
                        slices = None
                        rslt['slices'] = slices
                        half['slices'] = half['slices_half']

        angle_corr_slices = 0
        angle_corr_size = 0
        angle_corr_areas = 0
        angle_corr_airgap = False

        for size, n, rslt in check_rslt:
            areas = rslt['areas']
            slices = rslt['slices']
            size = rslt['areasize']

            if slices is None:  # ignore it
                continue

            delta_angle_corr = rslt.get('delta_corr', 0.0)
            # Angle Correction
            if with_angle_corr and self.delta_angle_corr is None:
                self.delta_angle_corr = delta_angle_corr
                angle_corr_slices = slices
                angle_corr_size = size
                angle_corr_areas = areas
                angle_corr_airgap = rslt.get('airgap', False)
            else:
                if with_angle_corr and self.delta_angle_corr != delta_angle_corr:
                    unsure_sym = True
                    if slices > angle_corr_slices and \
                       size > angle_corr_size * 0.5:
                        logger.debug("Angle Correction")
                        self.delta_angle_corr = delta_angle_corr
                        angle_corr_slices = slices
                        angle_corr_size = size
                        angle_corr_areas = areas
                        angle_corr_airgap = rslt.get('airgap', False)
                    elif angle_corr_airgap and \
                         slices >= angle_corr_slices and \
                         areas > angle_corr_areas and \
                         size > angle_corr_size * 0.2:
                        logger.debug("Angle Correction")
                        self.delta_angle_corr = delta_angle_corr
                        angle_corr_slices = slices
                        angle_corr_size = size
                        angle_corr_areas = areas
                        angle_corr_airgap = rslt.get('airgap', False)
            if slices:
                if start_delta is None:
                    start_delta = rslt['startdelta']
                    start_delta_corr = start_delta
                elif not rslt['startdelta'] != 0.0:
                    if start_delta_corr == 0.0:
                        start_delta_corr = rslt['startdelta']
                    elif not np.isclose(rslt['startdelta'], start_delta_corr,
                                        rtol=rtol, atol=atol):
                        slices = 0  # bad
                        rslt['slices'] = slices

            if slices is not None and slices <= 1:  # critical
                if areas <= max(1, max_areas / 5):
                    if size < max_size / 25:
                        slices = None
            if slices == 0:
                middles = rslt.get("middlelist", [])
                if self.check_missing_areas(middles,
                                            missing_middles):
                    logger.debug("Symmetry-Destroyer destroyed")
                    slices = None

            if rslt.get('upper_quarter', False):
                if size < max_size / 20:
                    slices = None

            if slices == 1:
                if rslt.get('upper_quarter', False):
                    slices = None

            if slices == 1:
                # symmetry killer
                if areas < max(2, max_areas / 6):
                    if size < max_size * 0.05:
                        slices = None  # ignore tiny areas

            parts_possible = self.calc_parts(parts_possible, slices)

        if unsure_sym:
            logger.warning("Warning: unsure symmetry")

        if parts_possible is None:
            parts_possible = 0
        return parts_possible, start_delta

    def check_missing_areas(self, middles, missing_middles):
        logger.debug("check_missing_areas")
        logger.debug(" -- mids = %s", middles)
        logger.debug(" -- missing mids = %s", missing_middles)
        if not missing_middles:
            return False
        if len(middles) == 0 or len(middles) > 2:
            return False

        for m in middles:
            mlist = [mm for mm, i in missing_middles
                     if np.isclose(m, mm, rtol=1e-3, atol=1e-3)]
            if not mlist:
                return False
        return True

    def calc_parts(self, parts1, parts2):
        logger.debug("Calc symmetry Parts (%s, %s)", parts1, parts2)
        if parts2 is None:
            logger.debug("return %s parts", parts1)
            return parts1
        if parts1 is None:
            logger.debug("return %s parts", parts2)
            return parts2
        if parts1 == 0 or parts2 == 0:
            logger.debug("return %s parts", 0)
            return 0
        parts = gcd(parts1, parts2)
        logger.debug("return %s parts", parts)
        return parts

    def create_cut_lines(self, parts, start_delta):
        self.geom.clear_cut_lines()
        for alpha in self.symmetry_lines(parts,
                                         self.startangle,
                                         start_delta,
                                         self.endangle):
            plus = self.geom.max_radius / 10
            min_radius = max(10, self.geom.min_radius - plus)
            p1 = point(self.geom.center, min_radius, alpha)
            p2 = point(self.geom.center, self.geom.max_radius + plus, alpha)
            line = Line(Element(start=p1, end=p2))
            line.init_attributes(color='green')
            self.geom.add_cut_line(line)

    def symmetry_lines(self, parts, startangle, start_delta, endangle):
        logger.debug("begin symmetry_lines from %s to %s with start %s",
                     startangle,
                     endangle,
                     start_delta)

        if less_equal(endangle, startangle):
            endangle += 2*np.pi

        sym = self.geom_part * parts
        delta = 2*np.pi/sym
        if start_delta == 0.0:
            if not is_same_angle(startangle, endangle):
                start_delta = delta

        sym_startangle = startangle + start_delta
        start = startangle + start_delta
        while less(start, endangle):
            yield start
            start += delta

        if is_same_angle(startangle, endangle):
            yield start

        # Damit man anschliessend ohne Umstände schneiden kann.
        self.set_symmetry_parameters(sym_startangle, parts, delta)
        logger.debug("end symmetry_lines")

    def set_symmetry_parameters(self, startangle, parts, delta):
        self.geom.sym_startangle = startangle
        self.geom.sym_endangle = startangle + delta
        self.geom.sym_slices = parts
        self.geom.sym_slice_angle = delta
        self.geom.sym_area = Area([], (0,0), 0.0)
        self.geom.sym_area.sym_startangle = self.geom.sym_startangle
        self.geom.sym_area.sym_endangle = self.geom.sym_endangle

    def check_symmetry_of_mirror(self, mirror_geom, mirrorangle):
        logger.debug("begin of Symmetry::check_symmetry_of_mirror")
        assert(mirror_geom is not None)

        axis_p = point(self.geom.center, self.geom.max_radius, mirrorangle)
        axis_m = line_m(self.geom.center, axis_p)
        axis_n = line_n(self.geom.center, axis_m)

        def counterpart_found(node, mirror_node, nodes, rtol, atol):
            hit_sloppy = 0
            for n in nodes:
                if points_are_close(mirror_node, n, rtol, atol):
                    logger.debug(" ---- %s is %s", node, mirror_node)
                    return 1, 1
                if points_are_close(round_point(mirror_node, 1),
                                    round_point(n, 1),
                                    rtol, atol):
                    logger.debug(" ++++ %s is %s", node, mirror_node)
                    hit_sloppy = 1

            logger.debug(" >>>> %s is NOT %s  (%s)",
                         node, mirror_node, hit_sloppy)
            return 0, hit_sloppy

        def check_differences(geom, mirror_geom):
            geom_ag_nodes = []
            geom_nodes = [n for n in geom.g.nodes() if not (n in geom_ag_nodes)]

            hits = 0
            hits_sloppy = 0
            for n in geom_nodes:
                mirror_n = mirror_point(n, geom.center, axis_m, axis_n)
                hit, hit_sloppy = counterpart_found(n, mirror_n,
                                                    mirror_geom.g.nodes(),
                                                    # self.rtol,
                                                    1e-3,
                                                    # self.atol):
                                                    1e-2)
                hits += hit
                hits_sloppy += hit_sloppy

            min_nodes = min(len(geom_nodes), int(len(geom_nodes) * 0.95) + 1)
            logger.debug("Nodes=%s,  Counterparts=%s (sloppy=%s)",
                         len(geom_nodes), hits, hits_sloppy)
            if hits < min_nodes:
                f = hits / len(geom_nodes)
            else:
                f = 1.0
            if hits_sloppy < min_nodes:
                f_sloppy = hits_sloppy / len(geom_nodes)
            else:
                f_sloppy = 1.0
            return f, f_sloppy

        # ----------------
        logger.debug("check geom - mirror")
        f1, f1_sloppy = check_differences(self.geom, mirror_geom)
        logger.debug("check mirror - geom")
        f2, f2_sloppy = check_differences(mirror_geom, self.geom)
        logger.debug("Factor 1: %s,  2: %s", f1, f2)
        if f1 >= 0.99 or f2 >= 0.99:
            ok = not (f1 < 0.9 or f2 < 0.9)
        else:
            ok = f1 > 0.97 and f2 > 0.97
        if not ok:
            if f1_sloppy > 0.97 and f2_sloppy > 0.97:
                logger.debug("  (A sloppy mirror found, but ignored)")
                ok = True
        logger.debug("end of Symmetry::check_symmetry_of_mirror => %s", ok)
        return ok



'''Calculate Magnet Losses with AMELA

'''
import sys
import json
import pathlib
import logging
from pathlib import Path
import femagtools.nc
import numpy as np
from numpy import pi, sin, cos
import subprocess

#  set logging
logger = logging.getLogger(__name__)


class Amela():
    '''Run Amela Calculation

    Args:
      workdir : working directory of femag calculation
        (The directory that includes the nc file)
      amela_dir: str (optional) amela directory
      magnet_data : calculation control name must be provided
        dict(name='test',
        -- the following parameters are optional
             mur=1.05,
             sigma=625000,
             hm=3,
             wm=20,
             lm=30,
             speed=1000,
             nsegwid=0,
             nseglen=0)
    '''

    def __init__(self, workdir: str, magnet_data: dict,
                 amela_dir=None):

        self.magn = magnet_data
        self.workdir = pathlib.Path(workdir)
        if amela_dir is not None:
            self.amela_dir = pathlib.Path(amela_dir)
        else:
            self.amela_dir = self.workdir

        self.jsonfile = []
        if sys.platform == 'win32':
            self.cmd = [str(self.amela_dir / 'AMELA.BAT')]
        else:
            self.cmd = [str(self.amela_dir / 'AMELA')]
        # default batch
        self.cmd.append('--b')
        # append calc options
        if 'speed' in self.magn:
            self.cmd.append(f"--speed {self.magn['speed']}")
        if 'nsegwid' in self.magn:
            self.cmd.append(f"--nsegwid {self.magn['nsegwid']}")
        if 'nseglen' in self.magn:
            self.cmd.append(f"--nseglen {self.magn['nseglen']}")

    def get_magnet_data_all(self, num_op):
        '''get all magnet data for all loadcases'''
        nc_name = self.workdir / self.magn['name']
        r = femagtools.nc.read(nc_name)
        return [r.get_magnet_data(ibeta=i) for i in num_op]

    def export_json(self, pm_data: list):
        '''Export magnet data to json files

        Args:
          pm_data: list of magnet data

        '''
        return
        pm_dir = self.amela_dir / self.magn['name']
        pm_dir.mkdir(exist_ok=True)
        for i in pm_data:
            filename = (pm_dir / i['name']).with_suffix('.json')
            self.jsonfile.append(str(filename))  # for the future use
            # pop out non necessary data
            i.pop('bl')
            i.pop('elcp')
            with filename.open('w') as f:
                json.dump(i, f)
            logger.info('Exporting %s ...', i['name'])

    def read_loss(self, pm_data: dict) -> dict:
        '''Read magnet losses

        Args:
          pm_data : dict

        Returns:
          losses
        '''
        losses = {}
        for i in range(len(pm_data)):
            dirname = self.magn['name'] + '/' + pm_data[i]['name'] + \
                '/EClosses.tab'
            result_name = self.amela_dir / dirname
            with result_name.open() as f:
                data = np.loadtxt(f)
                total_loss = np.mean(data[0:-1, -1])
                losses[pm_data[i]['name']] = {"loss_data":data, "total_loss": total_loss}
                logger.info("Magnet losses in superelement %s is %.3f W",
                            pm_data[i]['name'].split('se')[-1], total_loss)
        return losses

    def __call__(self, ialh=False) -> dict:
        '''Run amela calculation

        Args:
          ialh: use method IALH if True else 3DI

        Returns:
            losses
        '''
        # get magnet data
        nc_name = self.workdir / self.magn['name']
        nc = femagtools.nc.read(nc_name)
        num_cases = nc.el_fe_induction_1.shape[3] - 1
        if 'loadcase' in self.magn:
            indx = self.magn['loadcase'] - 1
        else:
            indx = num_cases
        if indx == 3:
            indx = num_cases  # avoid error

        r = nc.get_magnet_data(indx)
        # export to json
        #self.export_json(r)
        # run amela
        calc_method = 'IALH' if ialh else '3DI'
        cmd = self.cmd + ['--calc', calc_method, self.magn['name'] + '/']
        log_file = self.amela_dir / 'amela.out'
        logger.info("Calculating magnet losses with AMELA ...")
        with log_file.open('w') as output:
            subprocess.run(cmd, stdout=output)
        return self.read_loss(r)



import sys
import pathlib
import argparse
import logging
import logging.config
from femagtools.dxfsl.converter import convert

logger = logging.getLogger(__name__)

def main():
    argparser = argparse.ArgumentParser(
        description='Process SVG file and create a plot or FSL file.')
    argparser.add_argument('svgfile',
                           help='name of SVG file')
    argparser.add_argument('--rotor',
                           help='rotor without airgap in/out',
                           dest='rotor',
                           default='')
    argparser.add_argument('--stator',
                           help='stator without airgap in/out',
                           dest='stator',
                           default='')
    argparser.add_argument('--EESM',
                           help='Electric Excited Synchronous Motor',
                           dest='EESM',
                           action="store_true",
                           default=False)
    argparser.add_argument('-p', '--plot',
                           help='show plots',
                           dest='show_plots',
                           action="store_true")
    argparser.add_argument('--areas',
                           help='show all areas',
                           dest='show_areas',
                           action="store_true")
    argparser.add_argument('--id',
                           help='write id of areas',
                           dest='write_id',
                           action="store_true")
    argparser.add_argument('-f', '--fsl',
                           help='create fsl',
                           dest='write_fsl',
                           action="store_true")
    argparser.add_argument('-v', '--view',
                           help='show a view only',
                           dest='view',
                           action="store_true")
    args = argparser.parse_args()

    loglevel = logging.INFO
    logging.basicConfig(level=loglevel,
                        format='%(asctime)s %(message)s',
                        filemode='w')
    part = ()
    if args.stator:
        if args.rotor:
            logger.error("Stator or Rotor expected")
            sys.exit(1)
        part = ('stator', args.stator)
    elif args.rotor:
        part = ('rotor', args.rotor)

    if not args.write_fsl:
        if not (args.show_plots or args.show_areas or args.view):
            args.write_fsl = True

    res = convert(args.svgfile,  # SVG-Filename
                  part=part,
                  EESM=args.EESM,
                  view_only=args.view,
                  show_plots=args.show_plots,
                  show_areas=args.show_areas,
                  write_fsl=args.write_fsl,
                  write_id=args.write_id)

    if args.write_fsl:
        if res is not None:
            p = pathlib.Path(args.svgfile)
            basename = p.stem
            pathlib.Path(basename + '.fsl').write_text('\n'.join(res['fsl']))

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
  femagtools.dxfsl.concat
  ~~~~~~~~~~~~~~~~~~~~~~~

  Authors: Ronald Tanner, Beat Holm
"""
from __future__ import print_function

import numpy as np
import logging
import sys
from femagtools.dxfsl.shape import Element, Shape, Circle, Arc, Line
from femagtools.dxfsl.shape import is_Circle, is_Arc, is_Line
from femagtools.dxfsl.functions import points_are_close, positive_angle, elevation_angle
from femagtools.dxfsl.functions import Timer, alpha_line
from femagtools.dxfsl.journal import getJournal
import io
import time

logger = logging.getLogger('femagtools.concat')


#############################
#           concat          #
#############################


class Concatenation(object):
    def __init__(self,
                 rtol=1e-04,
                 atol=1e-04,
                 ndec=6,
                 mdec=1):
        self.rtol = rtol
        self.atol = atol
        self.ndec = ndec
        self.mdec = mdec
        self.line_y0 = Line(Element(start=(-1e+8, 0.0), end=(1e+8, 0.0)))
        self.line_x0 = Line(Element(start=(0.0, -1e+8), end=(0.0, 1e+8)))
        self.journal = getJournal()

        logger.debug("Concatenation(rtol=%s, atol=%s, ndec=%s)", rtol, atol, ndec)

    def __str__(self):
        return "rtol: {}\n".format(self.rtol) + \
               "atol: {}\n".format(self.atol) + \
               "ndec: {}\n".format(self.ndec)

    def concatenate_line_elements(self,
                                  src_elements,
                                  dest_elements):
        if len(src_elements) < 2:
            return 0

        src_list = []
        for e in src_elements:
            if not e.has_attribute("del"):
                e_p1, ep_2 = e.points_sorted()
                src_list.append((e_p1, e))
        if len(src_list) < 2:
            return 0

        logger.debug("begin of concatenate_line_elements (%s elements)",
                     len(src_list))

        src_list.sort()
        e1_p1, e1 = src_list[0]
        new_elements = []
        count = 0
        for e2_p1, e2 in src_list[1:]:
            el_new = e1.concatenate(None, None, e2,
                                    rtol=1e-06,
                                    atol=1e-06,
                                    mdec=self.mdec,
                                    overlapping=True)
            if el_new:
                logger.debug("NEW %s from %s to %s",
                             el_new.classname(),
                             el_new.p1,
                             el_new.p2)
                el_new.set_attribute("new")
                e1.set_attribute("del")
                e2.set_attribute("del")
                logger.debug("OLD %s from %s to %s",
                             e1.classname(),
                             e1.p1,
                             e1.p2)
                logger.debug("OLD %s from %s to %s",
                             e2.classname(),
                             e2.p1,
                             e2.p2)
                el_p1, el_p2 = el_new.points_sorted()
                dest_elements.append(el_new)
                src_elements.append(el_new)
                e1 = el_new
                count += 1
            else:
                e1 = e2

        logger.debug("end of concatenate_line_elements: %s concatenations", count)
        return count

    def concatenate_arc_elements(self,
                                 src_elements,
                                 dest_elements):
        if len(src_elements) < 2:
            return 0
        
        src_list = [e for e in src_elements if not e.has_attribute("del")]
        if len(src_list) < 2:
            return 0

        e1 = src_list[0]
        new_elements = []
        count = 0
        for e2 in src_list[1:]:
            el_new = e1.concatenate(None, None, e2,
                                    rtol=1e-06,
                                    atol=1e-06,
                                    mdec=self.mdec,
                                    overlapping=True)
            if el_new:
                logger.debug("NEW %s from %s to %s",
                             el_new.classname(),
                             el_new.p1,
                             el_new.p2)
                el_new.set_attribute("new")
                e1.set_attribute("del")
                e2.set_attribute("del")
                logger.debug("OLD %s from %s to %s",
                             e1.classname(),
                             e1.p1,
                             e1.p2)
                logger.debug("OLD %s from %s to %s",
                             e2.classname(),
                             e2.p1,
                             e2.p2)
                dest_elements.append(el_new)
                src_elements.append(el_new)
                e1 = el_new
                count += 1
            else:
                e1 = e2
        return count

    def concatenate_matching_line_elements(self,
                                           elements,
                                           only_tiny=False):
        logger.debug("Begin of concatenate_matching_line_elements")
        timer = Timer(start_it=True)

        if only_tiny:
            tiny = [e for e in elements if is_Line(e) and e.has_attribute("tiny")]
            if not tiny:
                timer.stop("-- no tiny lines found in %0.4f seconds --")
                return 0  # no tiny line exists
        
        def axis_point(e, dec=1):
            if np.isclose(e.p1[0], e.p2[0]):
                return (round(e.p1[0], dec), 0.0)
            if np.isclose(e.p1[1], e.p2[1]):
                return (0.0, round(e.p1[1], dec))
            px = e.intersect_line(self.line_y0, all=True)
            py = e.intersect_line(self.line_x0, all=True)
            if not (px or py):
                logger.error("axis_point: FATAL ERROR !!")
                return (0.0, 0.0)

            if not px:
                py = round(py[0][1], dec)
                return (0.0, py)
            if not py:
                px = round(px[0][0], dec)
                return (px, 0.0)
            
            px = round(px[0][0], dec)
            py = round(py[0][1], dec)
            
            if abs(px) > abs(py):
                return (0.0, py)
            return (px, 0.0)

        def elevation(e):
            return round(elevation_angle(alpha_line(e.p1, e.p2)), 2)

        def axis_points_are_close(a1, a2):
            #return a1 == a2
            if a1[0]:
                d = abs(a1[0] - a2[0])  # x
            else:
                d = abs(a1[1] - a2[1])  # y
            if d < 0.2:
                logger.debug("axis_points_are_close: %s - %s = %s", a1, a2, d)
            return d < 0.2

        elmts = [(round(e.m(999999.0), self.mdec),
                  axis_point(e, dec=4),
                  e.p1,
                  e) for e in elements if is_Line(e)]
        if not elmts:
            return 0  # no lines available

        elmts.sort()

        logger.debug("Concatenate Line Elements")
        #for m, a, p, e in elmts:
        #    logger.debug("Line from %s to %s [m=%s, a=%s]",
        #                 e.p1, e.p2, m, a)
        logger.debug("*************************")

        lines_available = len(elmts)
        count = 0
        if len(elmts) < 2:
            timer.stop()
            return 0

        new_elements = []
        m1, a1, p1, e1 = elmts[0]
        line_elements = [e1]
        for m2, a2, p2, e2 in elmts[1:]:
            if m1 == m2 and axis_points_are_close(a1, a2):
                line_elements.append(e2)
            else:
                if len(line_elements) > 1:
                    logger.debug("Begin Concatenate %s Lines: m=%s, a=%s",
                                 len(line_elements), m1, a1)
                    for l in line_elements:
                        logger.debug("-- %s", l)
                    line_count = self.concatenate_line_elements(line_elements, elements)
                    while line_count > 0:
                        count += line_count
                        line_count = self.concatenate_line_elements(line_elements, elements)
                    logger.debug("End Concatenate Lines: m=%s, a=%s", m1, a1)

                m1 = m2
                a1 = a2
                line_elements = [e2]

        logger.debug("Begin Concatenate Lines: m=%s, a=%s", m1, a1)
        line_count = self.concatenate_line_elements(line_elements, elements)
        while line_count > 0:
            count += line_count
            line_count = self.concatenate_line_elements(line_elements, elements)
        logger.debug("End Concatenate Lines: m=%s, a=%s", m1, a1)

        timer.stop("-- {} lines concatenated in %0.4f seconds --".format(count))
        line_list = [e for e in elements
                    if not e.has_attribute("del") and is_Line(e)]
        logger.debug("== %s of %s Line-Elements are remaining ==",
                     len(line_list),
                     lines_available)
        logger.debug("End of concatenate_matching_line_elements")
        return count

    def concatenate_matching_arc_elements(self,
                                          elements,
                                          only_tiny=False):
        logger.debug("Begin of concatenate_matching_arc_elements")
        timer = Timer(start_it=True)

        if only_tiny:
            tiny = [e for e in elements if is_Arc(e) and e.has_attribute("tiny")]
            if not tiny:
                timer.stop("-- no tiny arcs found in %0.4f seconds --")
                logger.debug("End of concatenate_matching_arc_elements")
                return 0  # no tiny arc exists

        def center_rounded(c):
            return (round(c[0], 6),round(c[1], 6))

        elmts = [(center_rounded(e.center),
                  round(e.radius, 6),
                  positive_angle(e.startangle),
                  e) for e in elements if is_Arc(e) or is_Circle(e)]
        if not elmts:
            timer.stop("-- no arcs found in %0.4f seconds --")
            logger.debug("End of concatenate_matching_arc_elements")
            return 0  # no arcs available

        elmts.sort()

        logger.debug("Concatenate Arc Elements")
        #for c, r, a, e in elmts:
        #    logger.debug("%s from %s to %s [c=%s, r=%s, a=%s]",
        #                 e.classname(), e.p1, e.p2, c, r, a)
        logger.debug("*************************")
 
        arcs_available = len(elmts)
        count = 0

        c1, r1, a1, e1 = elmts[0]
        arc_elements = [e1]
        for c2, r2, a2, e2 in elmts[1:]:
            if points_are_close(c1, c2) and np.isclose(r1, r2):
                arc_elements.append(e2)
            else:
                logger.debug("Begin Concatenate %s Arcs: c=%s, r=%s",
                             len(arc_elements), c1, r1)
                arc_count = self.concatenate_arc_elements(arc_elements, elements)
                while arc_count > 0:
                    count += arc_count
                    arc_count = self.concatenate_arc_elements(arc_elements, elements)
                logger.debug("End Concatenate Arcs: c=%s, r=%s", c1, r1)

                c1 = c2
                r1 = r2
                arc_elements = [e2]

        logger.debug("Begin Concatenate Arcs: c=%s, r=%s", c1, r1)
        arc_count = self.concatenate_arc_elements(arc_elements, elements)
        while arc_count > 0:
            count += arc_count
            arc_count = self.concatenate_arc_elements(arc_elements, elements)
        logger.debug("End Concatenate Arcs: c=%s, r=%s", c1, r1)

        timer.stop("-- {} arcs concatenated in %0.4f seconds --".format(count))

        arc_list = [e for e in elements
                    if not e.has_attribute("del") and
                    (is_Arc(e) or is_Circle(e))]
        logger.debug("== %s of %s Arc-Elements are remaining ==",
                     len(arc_list),
                     arcs_available)
        logger.debug("End of concatenate_matching_arc_elements")
        return count

    def concatenate_matching_elements(self,
                                      src_elements,
                                      main=False,
                                      only_tiny=False):
        logger.debug("Begin of concatenate_matching_elements")
        timer = Timer(start_it=True)
        count_lines = count_arcs = 0
        count_lines = self.concatenate_matching_line_elements(src_elements,
                                                              only_tiny=only_tiny)
        count_arcs = self.concatenate_matching_arc_elements(src_elements,
                                                            only_tiny=only_tiny)
        if only_tiny:
            if count_lines == 0 and count_arcs == 0:
                #  no action
                timer.stop("-- no concatenations in %0.4f seconds --")
                return False, src_elements

        count = count_lines + count_arcs
        if main:
            self.journal.put_concat_lines(count_lines)
            self.journal.put_concat_arcs(count_arcs)

        new_list = [e for e in src_elements if not e.has_attribute("del")]

        t = timer.stop("-- {} concatenations in %0.4f seconds --".format(count))
        if main:
            self.journal.put('time_concatenation', t)

        logger.debug("End of concatenate_matching_elements")
        return count, new_list



"""Read FEMAG vtu files

"""
import logging
import pathlib
import numpy as np
import vtk
from vtkmodules.util.numpy_support import vtk_to_numpy


class Reader(object):
    '''read vtu-files

        Args:
          pathname : str Directory of result files (vtu-files) or a single vtu file

    '''

    def __init__(self, pathname):
        self.data = {}

        self.reader = vtk.vtkXMLUnstructuredGridReader()
        self.output = self.reader.GetOutput()

        self.field_data_names = []
        self.point_data_names = []
        self.cell_data_names = []
        assert pathlib.Path(pathname).exists(), f"{pathname} not found"
        if pathlib.Path(pathname).suffix == '.vtu':
            self.filenames = [pathlib.Path(pathname)]
        else:
            self.filenames = sorted(pathlib.Path(pathname).glob(
                "*.vtu"))

        self.reader.SetFileName(str(self.filenames[0]))
        self.reader.Update()
        self.field_data_names = [
            self.output.GetFieldData().GetAbstractArray(i).GetName()
            for i in range(self.output.GetFieldData().GetNumberOfArrays())]
        self.point_data_names = [
            self.output.GetPointData().GetAbstractArray(i).GetName()
            for i in range(self.output.GetPointData().GetNumberOfArrays())]
        self.cell_data_names = [
            self.output.GetCellData().GetAbstractArray(i).GetName()
            for i in range(self.output.GetCellData().GetNumberOfArrays())]

        self.set_time_window(0.0, 0.0)

    def get_data_names(self):
        '''Read the list of values stored in the vtu files

        Returns:
            List of values stored in the vtu files

        '''
        return (self.field_data_names +
                self.point_data_names +
                self.cell_data_names)

    def read_data(self, data_list):
        '''Extracts data from the vtu files

        Args:
          data_list : fist of str List of values to extract from vtu_files

        '''
        for data_name in data_list:
            if data_name in self.field_data_names:
                self.data[data_name] = []
            elif data_name in self.point_data_names:
                self.data[data_name] = []
            elif data_name in self.cell_data_names:
                self.data[data_name] = []
            else:
                raise Exception('unknown data name "' + data_name+'"')

        for filename in self.filenames:
            self.reader.SetFileName(str(filename))
            self.reader.Update()

            for data_name in data_list:
                if data_name in self.field_data_names:
                    self.data[data_name].append(
                        self.output.GetFieldData().GetAbstractArray(data_name))
                if data_name in self.point_data_names:
                    self.data[data_name].append(
                        self.output.GetPointData().GetAbstractArray(data_name))
                if data_name in self.cell_data_names:
                    self.data[data_name].append(
                        self.output.GetCellData().GetAbstractArray(data_name))

        return "done"

    def set_time_window(self, start, end):
        '''Set time window

        Args
          start: float Start of the time window
        end: float End of the time window

        Only values within the time window are output by the functions
            get_field_vector
            get_point_vector
            get_cell_vector
            get_data_vector
        At start = 0.0 the values are read out starting from the first value
        At end = 0.0 the values are read out up to the last value

        '''
        try:
            if "time [s]" not in self.data:
                self.read_data(['time [s]'])

            if start == 0 or start <= self.data['time [s]'][0].GetValue(0):
                self.istart = 0
            else:
                self.istart = 0
                for i in range(len(self.data['time [s]'])):
                    if start >= self.data['time [s]'][i].GetValue(0):
                        self.istart = i+1
            if end == 0 or end >= self.data['time [s]'][-1].GetValue(0):
                self.iend = len(self.data['time [s]'])
            else:
                self.iend = 0
                for i in range(len(self.data['time [s]'])):
                    if end <= self.data['time [s]'][i].GetValue(0):
                        self.iend = i
        except:
            self.istart = None
            self.iend = None

    def get_field_vector(self, field_data) -> list:
        '''Read field data

        Args:
          field_data : str Name of field to read

        Returns:
            List of field values within the time window
        '''
        if field_data not in self.data:
            self.read_data([field_data])

        if self.istart:
            start = self.istart
        else:
            start = 0
        if self.iend:
            end = self.iend
        else:
            end = len(self.data[field_data])

        field_vec = []
        # for i in range(self.istart,self.iend):
        for i in range(start, end):
            field_vec.append(self.data[field_data][i].GetValue(0))
        return field_vec

    # pnt = node-key, >0
    def get_point_vector(self, pnt_data, pnt) -> list:
        '''Read point data

        Args:
          point_data : str Name of field to read
          pnt : int Key of point

        Returns:
          List of point values within the time window

        '''
        if pnt_data not in self.data:
            self.read_data([pnt_data])

        if self.istart:
            start = self.istart
        else:
            start = 0
        if self.iend:
            end = self.iend
        else:
            end = len(self.data[pnt_data])

        point_vec = []
        for i in range(start, end):
            point_vec.append(self.data[pnt_data][i].GetValue(pnt-1))
        return point_vec

    def get_cell_vector(self, cell_data, cell=0) -> list:
        '''Read cell data

        Args:
          cell_data : str  Name of field to read
          cell : int Key of cell

        Returns:
            List of cell values within the time window

        '''
        if cell_data not in self.data:
            self.read_data([cell_data])

        if cell<=0:
            return vtk_to_numpy(
                self.output.GetCellData().GetAbstractArray(cell_data))
        i = self.cell_data_names.index(cell_data)
        noc = self.output.GetCellData().GetAbstractArray(i).GetNumberOfComponents()
        if noc == 1:
            cell_vec = []
        else:
            cell_vec_x = []
            cell_vec_y = []
            cell_vec_z = []

        if self.istart:
            start = self.istart
        else:
            start = 0
        if self.iend:
            end = self.iend
        else:
            end = int(len(self.data[cell_data]))

        for i in range(start, end):
            if noc == 1:
                cell_vec.append(self.data[cell_data][i].GetValue(cell-1))
            else:
                cell_vec_x.append(
                    self.data[cell_data][i].GetValue(noc*(cell-1)))
                cell_vec_y.append(
                    self.data[cell_data][i].GetValue(noc*(cell-1)+1))
                cell_vec_z.append(
                    self.data[cell_data][i].GetValue(noc*(cell-1)+2))
        if noc == 1:
            return cell_vec
        else:
            return [cell_vec_x, cell_vec_y, cell_vec_z]

    def hrphi(self, scf, elements, magtemp):
        """return H values for each element in polar coord system
        (experimental)
        Args:
          elements: list of model elements
          magtemp: magnet temperature in degree Celsius
          scf: scale factor (poles/poles in model)
        """
        MUE0 = 4e-7*np.pi
        b = vtk_to_numpy(
            self.output.GetCellData().GetAbstractArray("b"))[:, :2]
        e = elements[0]
        x, y = np.mean(vtk_to_numpy(
            self.output.GetCell(e.key-1).GetPoints().GetData()), axis=0)[:2]
        rotpos = np.arctan2(y, x) - np.arctan2(e.center[1], e.center[0])
        psign = 1 if rotpos / scf > 1 else -1

        h = np.empty((len(elements), 2))
        for i, e in enumerate(elements):
            x, y = e.center
            alfa = np.arctan2(y, x)
            b1, b2 = b[e.key-1]
            btempc = 1. + e.br_temp_coef*(magtemp - 20.)
            m1 = e.mag[0]*np.cos(alfa) + e.mag[1]*np.sin(alfa)
            m2 = -e.mag[0]*np.sin(alfa) + e.mag[1]*np.cos(alfa)
            h[i] = abs(e.reluc[0])/MUE0*np.array((
                (b1 - psign*btempc*m1), (b2 - psign*btempc*m2)))
        logging.debug("H shape %s", h.shape)
        return h


    def demag(self, elements):
        """return demag values for each element in kA/m
        Args:
          elements: list of model elements
        Returns:
          list of demags from each vtu file in set
        """
        dlist = []
        for filename in self.filenames:
            self.reader.SetFileName(str(filename))
            self.reader.Update()
            d = vtk_to_numpy(
                self.output.GetCellData().GetAbstractArray("demagnetization"))
            dm = np.empty((len(elements)))
            for i, e in enumerate(elements):
                dm[i] = d[e.key-1]
            dlist.append(dm)
        logging.debug("Demag shape %s", np.array(dlist).shape)
        return dlist


    def get_data_vector(self, data_name, key=0) -> list:
        '''Read data of fiels, point or cell

        Args:
          data_name : str Name of data to read
          key : int (optional) Key of point or cell

        Returns:
            List of values within the time window
        '''
        if data_name in self.field_data_names:
            return self.get_field_vector(data_name)
        if data_name in self.point_data_names:
            return self.get_point_vector(data_name, key)
        if data_name in self.cell_data_names:
            return self.get_cell_vector(data_name, key)
        return []

    def __repr__(self):
        fmt = [f"N Cells {self.output.GetNumberOfCells()}  N Points {self.output.GetNumberOfPoints()}"]
        fmt += [
            f"{self.output.GetPointData().GetAbstractArray(i).GetName()}   Points"
            for i in range(self.output.GetPointData().GetNumberOfArrays())]
        fmt += [
            f"{self.output.GetCellData().GetAbstractArray(i).GetName()}   Cells"
            for i in range(self.output.GetCellData().GetNumberOfArrays())]
        return '\n'.join(fmt)


def read(filename) -> Reader:
    """
    Read vtu file and return Reader object.

    Args:
        filename: name of vtu file to be read
    """
    return Reader(filename)



#!/usr/bin/env python
#
import unittest
import os
import femagtools.bch
from io import open
import numpy as np


class BchReaderTest(unittest.TestCase):

    def read_bch(self, filename):
        testPath = os.path.join(os.path.split(__file__)[0], 'data')
        if len(testPath) == 0:
            testPath = os.path.join(os.path.abspath('.'), 'data')
        r = femagtools.bch.Reader()
        with open('{0}/{1}'.format(testPath, filename),
                  encoding='latin1') as f:
            r.read(f)
        return r

    def test_read_cogging(self):
        bch = self.read_bch('cogging.BATCH')
        self.assertEqual(bch.version, '7.9.147 November 2012')
        self.assertEqual(bch.nodes, 2315)
        self.assertEqual(bch.elements, 3305)
        self.assertEqual(bch.quality, 100.0)

        self.assertEqual(len(bch.torque_fft), 1)
        self.assertEqual(len(bch.torque_fft[0]), 5)
        self.assertTrue('order' in bch.torque_fft[0])
        self.assertTrue('torque' in bch.torque_fft[0])
        self.assertEqual(len(bch.torque_fft[0]['torque']), 5)
        self.assertEqual(bch.torque_fft[0]['order'], [4, 12, 24, 36, 48])

        self.assertEqual(sorted(bch.flux.keys()), ['1', '2', '3'])
        self.assertEqual(sorted(bch.flux['1'][0].keys()),
                         sorted(['displ', 'voltage_four',
                                 'current_k', 'flux_k',
                                 'voltage_ir', 'displunit',
                                 'voltage_dpsi']))
        self.assertEqual(len(bch.flux['1'][0]['flux_k']), 61)
        self.assertEqual(bch.flux_fft['1'][0]['order'], [1, 3, 5, 7, 9, 11])

        self.assertEqual(len(bch.torque), 1)
        self.assertEqual(sorted(bch.torque[0].keys()),
                         sorted(['angle', 'force_y', 'force_x', 'torque',
                                 'current_1', 'ripple', 't_idpsi']))
        self.assertEqual(len(bch.torque[0]['torque']), 61)

        self.assertAlmostEqual(bch.losses[0]['winding'], 0.0, 1)
        self.assertAlmostEqual(bch.losses[0]['stajo'], 0.458, 2)
        self.assertAlmostEqual(bch.losses[0]['staza'], 0.344, 3)
        self.assertAlmostEqual(bch.losses[0]['magnetJ'], 0.006, 3)
        # self.assertAlmostEqual(bch.losses[0]['rotfe'], 0.000, 3)

        self.assertAlmostEqual(bch.lossPar['fo'][0], 50.0, 1)
        self.assertAlmostEqual(bch.lossPar['fo'][1], 50.0, 1)
        self.assertEqual(bch.get(('machine', 'p')), 2)
        np.testing.assert_almost_equal(bch.inertia, [0.230195e-3, 0.011774e-3])

    def test_read_sctest(self):
        bch = self.read_bch('sctest.BATCH')

        self.assertEqual(len(bch.torque_fft), 1)
        self.assertEqual(len(bch.scData['ia']), 134)
        self.assertAlmostEqual(bch.scData['ikd'], 0.0, 1)
        self.assertAlmostEqual(bch.scData['iks'], 1263.581, 2)
        self.assertAlmostEqual(bch.scData['tks'], 1469.736, 2)

    def test_read_pmsim(self):
        bch = self.read_bch('pmsim.BATCH')

        self.assertEqual(len(bch.torque_fft), 2)
        self.assertTrue('order' in bch.torque_fft[0])
        self.assertTrue('torque' in bch.torque_fft[0])
        self.assertEqual(len(bch.torque_fft[0]['torque']), 7)
        self.assertEqual(bch.torque_fft[1]['order'], [0, 12, 24, 30, 36, 42])

        self.assertEqual(sorted(bch.flux['1'][0].keys()),
                         sorted(['displ', 'voltage_four',
                                 'current_k', 'flux_k',
                                 'voltage_ir', 'displunit',
                                 'voltage_dpsi']))
        self.assertEqual(len(bch.flux['1'][0]['flux_k']), 46)

        self.assertEqual(len(bch.torque), 2)
        self.assertTrue('torque' in bch.torque[1])
        self.assertEqual(len(bch.torque[1]['torque']), 46)

        self.assertTrue('ld' in bch.dqPar)
        self.assertAlmostEqual(bch.dqPar['i1'][1], 49.992, 3)
        self.assertAlmostEqual(bch.dqPar['ld'][0], 9.9e-3, 6)
        self.assertAlmostEqual(bch.dqPar['ld'][0], 9.9e-3, 6)
        self.assertAlmostEqual(bch.dqPar['u1'][1], 358.10, 2)
        self.assertAlmostEqual(bch.dqPar['torque'][0], 65.3, 1)

        self.assertAlmostEqual(bch.machine['i1'], 50.0)

        self.assertAlmostEqual(bch.lossPar['fo'][0], 50.0, 1)

        np.testing.assert_almost_equal(bch.losses[-1]['fft']['stajo']['order_el'],
                                       [1, 3, 5, 7, 9, 11, 13, 15])
        np.testing.assert_almost_equal(bch.losses[-1]['fft']['stajo']['freq'],
                                       [100.0, 300.0, 500.0, 700.0, 900.0,
                                        1100.0, 1300.0, 1500.0])
        np.testing.assert_almost_equal(bch.losses[-1]['fft']['stajo']['hyst'],
                                       [10.33, 9.391, 9.391, 9.391, 3.348,
                                        2.971, 1.476, 0.882])
        np.testing.assert_almost_equal(bch.losses[-1]['fft']['stajo']['eddy'],
                                       [15.804, 142.234, 395.094, 774.383,
                                        455.591, 603.881, 419.063, 333.395])

        np.testing.assert_almost_equal(bch.losses[-1]['fft']['staza']['order_el'],
                                       [1, 3, 5, 7, 9, 11, 13, 15])
        np.testing.assert_almost_equal(bch.losses[-1]['fft']['staza']['freq'],
                                       [100.0, 300.0, 500.0, 700.0, 900.0, 1100.0, 1300.0, 1500.0])
        np.testing.assert_almost_equal(bch.losses[-1]['fft']['staza']['hyst'],
                                       [8.641, 7.774, 7.774, 7.748, 3.679, 2.915, 1.303, 0.626])
        np.testing.assert_almost_equal(bch.losses[-1]['fft']['staza']['eddy'],
                                       [13.065, 117.587, 326.631, 637.999, 500.663, 592.805, 370.023, 236.594])

    def test_read_pmsim_9(self):
        bch = self.read_bch('pmsim-9.BATCH')

        self.assertAlmostEqual(bch.machine['plfe'][0], 2540.2, 1)
        self.assertAlmostEqual(bch.machine['plfe'][1], 2020.5, 1)
        self.assertAlmostEqual(bch.dqPar['up'][0], 259.4, 1)

        np.testing.assert_almost_equal(bch.losses[-1]['fft']['stajo']['order_mech'],
                                       [6, 18, 30, 42, 54, 90, 114])
        np.testing.assert_almost_equal(bch.losses[-1]['fft']['stajo']['order_el'],
                                       [1.0, 3.0, 5.0, 7.0, 9.0, 15.0, 19.0])
        np.testing.assert_almost_equal(bch.losses[-1]['fft']['stajo']['freq'],
                                       [400.0, 1200.0, 2000.0, 2800.0, 3600.0, 6000.0, 7600.0])
        np.testing.assert_almost_equal(bch.losses[-1]['fft']['stajo']['hyst'],
                                       [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
        np.testing.assert_almost_equal(bch.losses[-1]['fft']['stajo']['eddy'],
                                       [1637.884, 225.861, 93.969, 19.904, 6.661, 3.043, 1.752])

        assert [round(l*1e3, 4) for l in bch.dqPar['Lho']] == [0.5908, 0.6583]

    def test_read_relsim(self):
        bch = self.read_bch('relsim.BATCH')

        self.assertEqual(len(bch.torque), 1)
        self.assertTrue('torque' in bch.torque[0])
        self.assertAlmostEqual(np.mean(bch.torque[0]['torque']), 5.656, 2)
        self.assertAlmostEqual(bch.dqPar['u1'][1], 274.5, 1)
        self.assertAlmostEqual(bch.dqPar['torque'][0], 5.775, 1)

        np.testing.assert_almost_equal(bch.losses[-1]['fft']['stajo']['freq'],
                                       [50.0])
        np.testing.assert_almost_equal(bch.losses[-1]['fft']['stajo']['hyst'],
                                       [0.152])
        np.testing.assert_almost_equal(bch.losses[-1]['fft']['staza']['hyst'],
                                       [0.066])

    def test_read_pmsim_external(self):
        bch = self.read_bch('pmsim-external.BATCH')

        self.assertTrue('ld' in bch.dqPar)
        self.assertAlmostEqual(bch.dqPar['i1'][1], 49.992, 3)
        self.assertAlmostEqual(bch.dqPar['ld'][0], 0.86688e-3, 6)
        self.assertAlmostEqual(bch.dqPar['ld'][0], 0.86688e-3, 6)
        self.assertAlmostEqual(bch.dqPar['u1'][1], 2394.0, 2)
        self.assertAlmostEqual(bch.dqPar['torque'][0], 1137.92, 1)

        np.testing.assert_almost_equal(bch.losses[-1]['fft']['stajo']['order_el'],
                                       [1, 3])
        np.testing.assert_almost_equal(bch.losses[-1]['fft']['stajo']['freq'],
                                       [800.0, 2400.0])
        np.testing.assert_almost_equal(bch.losses[-1]['fft']['stajo']['hyst'],
                                       [2619.555, 49.438])
        np.testing.assert_almost_equal(bch.losses[-1]['fft']['stajo']['eddy'],
                                       [15512.529, 1186.523])

        np.testing.assert_almost_equal(bch.losses[-1]['fft']['staza']['order_el'],
                                       [1, 3, 5])
        np.testing.assert_almost_equal(bch.losses[-1]['fft']['staza']['freq'],
                                       [800.0, 2400.0, 4000.0])
        np.testing.assert_almost_equal(bch.losses[-1]['fft']['staza']['hyst'],
                                       [5688.175, 296.19, 0.989])
        np.testing.assert_almost_equal(bch.losses[-1]['fft']['staza']['eddy'],
                                       [43864.352, 7108.561, 39.563])

    def test_read_psidq(self):
        bch = self.read_bch('psidpsiq.BATCH')

        self.assertEqual(len(bch.torque_fft), 10)
        self.assertTrue('order' in bch.torque_fft[0])
        self.assertTrue('torque' in bch.torque_fft[0])
        self.assertEqual(len(bch.torque_fft[0]['torque']), 7)
        self.assertEqual(bch.torque_fft[0]['order'],
                         [0, 4, 8, 12, 16, 20, 24])

        self.assertEqual(sorted(bch.flux.keys()), ['1', '2', '3'])
        self.assertEqual(len(bch.flux['1']), 10)
        self.assertTrue('flux_k' in bch.flux['1'][0])
        self.assertEqual(len(bch.flux['1'][0]['flux_k']), 16)

        self.assertEqual(len(bch.torque), 10)
        self.assertEqual(len(bch.torque[-1]['torque']), 16)

        self.assertEqual(len(bch.psidq), 7)
        self.assertEqual(len(bch.psidq_ldq), 6)
        self.assertEqual(len(bch.psidq['psid']), 3)
        self.assertEqual(len(bch.psidq_ldq['ld']), 3)

        self.assertEqual(len(bch.psidq['losses']), 17)
        self.assertEqual(len(bch.psidq['losses']['styoke']), 3)
        self.assertTrue('id' in bch.airgapInduction)
        self.assertEqual(bch.airgapInduction['id'],
                         [-200.0, -100.0, 0.0])
        self.assertEqual(len(bch.airgapInduction['Ba']), 3)
        self.assertEqual(len(bch.airgapInduction['Bm'][0]), 3)

    def test_read_ldq(self):
        bch = self.read_bch('ldq.BATCH')
        self.assertEqual(len(bch.torque_fft), 13)
        self.assertTrue('order' in bch.torque_fft[0])
        self.assertTrue('torque' in bch.torque_fft[0])
        self.assertEqual(len(bch.torque_fft[0]['torque']), 8)
        self.assertEqual(bch.torque_fft[0]['order'], [12, 36, 48, 56, 60,
                                                      72, 76, 84])

        self.assertEqual(sorted(bch.flux.keys()), ['1', '2', '3'])
        self.assertEqual(len(bch.flux['1']), 13)
        self.assertEqual(len(bch.flux['1'][0]), 7)
        self.assertTrue('flux_k' in bch.flux['1'][0])
        self.assertEqual(len(bch.flux['1'][0]['flux_k']), 46)

        self.assertEqual(len(bch.torque), 13)
        self.assertEqual(len(bch.torque[-1]['torque']), 46)

        self.assertEqual(len(bch.ldq['losses']), 7)
        self.assertEqual(len(bch.ldq['losses']['styoke']), 4)

        self.assertTrue('i1' in bch.airgapInduction)
        self.assertEqual(len(bch.airgapInduction['i1']), 3)
        self.assertEqual(len(bch.airgapInduction['an']), 4)
        self.assertEqual(len(bch.airgapInduction['an'][0]), 4)

    def test_read_pmsim2(self):
        bch = self.read_bch('PM_270_L8_001.BATCH')
        self.assertAlmostEqual(bch.dqPar['i1'][1], 70.0, 1)
        self.assertAlmostEqual(bch.dqPar['beta'][0], -38.0, 1)

    def test_read_linearforce(self):
        bch = self.read_bch('linearForce.BATCH')

        self.assertEqual(len(bch.linearForce), 1)
        self.assertEqual(len(bch.linearForce[0]['displ']), 26)
        self.assertEqual(bch.linearForce[0]['displ'][5], 15.0)
        self.assertEqual(bch.linearForce[0]['force_x'][7], -0.3439)
        self.assertEqual(bch.linearForce[0]['force_y'][2], 03107.0)
        self.assertEqual(bch.linearForce[0]['magnet_1'][13], 10.0)
        self.assertEqual(bch.linearForce_fft[0]['force'][0], 0.3483)
        self.assertEqual(bch.linearForce_fft[1]['force'][0], 3157.)

        self.assertEqual(len(bch.linearForce_fft), 2)
        self.assertEqual(len(bch.flux_fft), 3)

    def test_read_linmot_z(self):
        bch = self.read_bch('linmot_z.BATCH')
        self.assertEqual(len(bch.linearForce), 2)
        self.assertEqual(max(bch.linearForce[1]['force_z']), 4074.0)

    def test_dq(self):
        bch = self.read_bch('dq.BATCH')
        bch.get(['torque', 'torque']) == []
        bch.get(['linearForce[-1]', 'ripple_x']) == 0.0
        assert bch.get(['linearForce', 'ripple_z']) is None
#        self.assertAlmostEqual(bch.dqPar['psid'][0], 2.7294321753800737, 5)
#        self.assertAlmostEqual(bch.dqPar['psiq'][0], 1.0899999999999999, 5)
        self.assertAlmostEqual(bch.dqPar['psid'][0], 1.93, 5)
        self.assertAlmostEqual(bch.dqPar['psiq'][0], 0.77074639149333668, 5)

    def test_read_felosses(self):
        bch = self.read_bch('rel-felosses.BATCH')
        self.assertEqual(len(bch.losses), 4)
        self.assertEqual(bch.losses[-1]['stajo'], 4425.106)
        self.assertEqual(bch.losses[-1]['staza'], 7504.659)

    def test_read_pmsim_demag(self):
        bch = self.read_bch('PMREL-4p-skewed.BATCH')

        self.assertEqual(len(bch.demag), 9)
        self.assertEqual([-370.92, -2241.79, -2236.31],
                         [d['H_max'] for d in bch.demag if d['segment'] == 3])

    def test_read_characteristics(self):
        bch = self.read_bch('char.BATCH')
        self.assertEqual(len(bch.characteristics), 1)
        self.assertEqual(len(bch.characteristics[0].keys()), 19)
        self.assertEqual(len(bch.characteristics[0]['speed_torque']['n']), 16)

    def test_read_asterisks(self):
        bch = self.read_bch('PM-with-asterisks_001.BATCH')
        self.assertTrue(np.isnan(bch.nodes))
        self.assertAlmostEqual(bch.airgapInduction['an'][0][8][0], 0.0690, 1)
        self.assertAlmostEqual(bch.airgapInduction['an'][0][9][0], -0.9915, 1)

    def test_read_dist_leak(self):
        bch = self.read_bch('PM-4p-distleak.BATCH')
        self.assertTrue(bch.leak_dist_wind)
        self.assertEqual(bch.leak_dist_wind['nseg'], 4)

    def test_read_felosses_ldq_b2(self):
        bch = self.read_bch('ldqlosses-2024b2.BATCH')
        self.assertEqual(bch.ldq['losses']['styoke_excess'][0],  [1.45, 2.114, 2.175, 2.175])
        self.assertEqual(bch.ldq['losses']['stteeth_excess'][0],  [4.309, 4.973, 5.034, 5.034])
        self.assertEqual(bch.ldq['losses']['rotor_excess'][0],  [0.0, 0.0, 0.0, 0.7933])

    def test_read_felosses_ldq_b2(self):
        bch = self.read_bch('psdqlosses-2024b2.BATCH')
        self.assertEqual(bch.psidq['losses']['styoke_excess'][0],  [2.235, 2.235, 2.839])
        self.assertEqual(bch.psidq['losses']['stteeth_excess'][0],  [6.58, 6.576, 7.176])
        self.assertEqual(bch.psidq['losses']['rotor_excess'][0],  [0.0, 0.0, 0.0])

    def test_read_pm_sym_outer_rotor(self):
        bch = self.read_bch('pm_sym_fast_outer_rotor.BATCH')
        self.assertEqual(bch.losses[-1]['staza'],  54.219)
        self.assertEqual(bch.losses[-1]['stajo'],  8.853)
        self.assertEqual(bch.losses[-1]['rotfe'],  0.091)
        self.assertEqual(list(bch.losses[-1]['fft']['rotor']['order_mech']),  [12, 24])
        self.assertEqual(list(bch.losses[-1]['fft']['rotor']['eddy']),  [0.057, 0.033])
        self.assertEqual(list(bch.losses[-1]['fft']['staza']['order_mech']),  [4, 12, 20, 28, 36, 44, 52])
        self.assertEqual(list(bch.losses[-1]['fft']['staza']['eddy']),  [45.215, 4.582, 2.881, 0.839, 0.425, 0.197, 0.059])
        self.assertEqual(list(bch.losses[-1]['fft']['stajo']['order_mech']),  [4, 12, 20, 28, 36])
        self.assertEqual(list(bch.losses[-1]['fft']['stajo']['eddy']),  [7.845, 0.64, 0.311, 0.044, 0.011])

    def test_read_ldlq_outer_rotor(self):
        bch = self.read_bch('ldlq_outer_rotor.BATCH')
        self.assertEqual(bch.ldq['losses']['styoke_eddy'][0],  [63.32, 39.69, 42.85, 65.62])
        self.assertEqual(bch.ldq['losses']['styoke_hyst'][0],  [0.0, 0.0, 0.0, 0.0])
        self.assertEqual(bch.ldq['losses']['rotor_eddy'][0],  [0.05937, 0.4195, 1.74, 2.705])
        self.assertEqual(bch.ldq['losses']['rotor_hyst'][0],  [0.0, 0.0, 0.0, 0.0])

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
  femagtools.dxfsl.areabuilder
  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  Authors: Ronald Tanner, Beat Holm
"""
from __future__ import print_function

import numpy as np
import logging
import sys
from femagtools.dxfsl.shape import Element, Shape, Circle, Arc, Line
from femagtools.dxfsl.shape import is_Circle, is_Arc, is_Line
from femagtools.dxfsl.area import Area, TYPE_AIR
from femagtools.dxfsl.functions import points_are_close, nodes_are_equal, distance
from femagtools.dxfsl.functions import normalise_angle, positive_angle, point
from femagtools.dxfsl.functions import alpha_line, alpha_points, alpha_angle
from femagtools.dxfsl.functions import less, less_equal, greater, is_same_angle
from femagtools.dxfsl.functions import Timer
from femagtools.dxfsl.journal import getJournal
import io
import time

logger = logging.getLogger('femagtools.areabuilder')
original_log_level = None

def disable_logging():
    logger
    global original_log_level
    # Store the current log level to restore it later
    original_log_level = logger.getEffectiveLevel()
    logger.debug("Logging level %s disabled", original_log_level)

    # Set the log level to a higher level, e.g., WARNING or CRITICAL
    logging.disable(logging.ERROR)


def enable_logging():
    #if original_log_level is not None:
    # Restore the original log level after the tests
    logging.disable(logging.NOTSET)
    logger.debug("Logging level %s enabled", original_log_level)

    
def log_lefthand(left, edge1, edge2):
    if left:
        edge1.log_edge("**** left")
    else:
        edge2.log_edge("**** left")


#############################
#        areabuilder        #
#############################

class EdgeInfo(object):
    def __init__(self,
                 edge_data=None,
                 n1 = None,
                 n2 = None,
                 handmade=False):
        assert(edge_data)
        assert(n1)
        assert(n2)

        if not handmade:
            e = edge_data.get('object', None)
            if not e:
                raise ValueError("Fatal: no object in edge_data %s", edge_data)          
            x = e.get_node_number(n1)
        else:
            e = None
            x = 0
            
        self.n1 = n1
        self.n2 = n2
        self.data = edge_data
        self.element = e
        self.x = x
        self.tracked = edge_data.get(x, False)
        self.alpha = 0.0
        self.startangle = None
        self.angle = None
        self.name = None

    def log_edge(self, text):
        logger.debug("%s: %s[%s] from %s to %s [%s]",
                     text,
                     self.classname(),
                     self.name,
                     self.n1, self.n2,
                     self.angle)

    def log_name(self, name):
        self.name = name

    def get_reverse_edge(self):
        return EdgeInfo(edge_data=self.data, n1=self.n2, n2=self.n1)

    def is_circle(self):
        return is_Circle(self.element)

    def is_arc(self):
        return is_Arc(self.element)

    def is_line(self):
        return is_Line(self.element)

    def classname(self):
        return self.element.classname()

    def is_start_edge(self):
        return self.startangle is not None

    def set_tracked(self):
        self.data[self.x] = True  # footprint
        self.tracked = True

    def set_all_tracked(self):
        self.data[1] = True  # footprint
        self.data[2] = True  # footprint
        self.tracked = True

    def is_tracked(self):
        return self.tracked

    def set_start_angle(self):
        start_angle = positive_angle(self.n2_angle_ingoing())
        self.startangle = start_angle

    def set_direction_angle(self, start_edge):
        start_angle = start_edge.startangle
        logger.debug("begin set_direction_angle: of %s", self.name)
        self.angle = positive_angle(alpha_angle(start_angle, self.n1_angle_ingoing()))
        logger.debug("set_direction_angle: angle is %s", self.angle)

        if is_same_angle(0.0, self.angle, atol=0.015):  # 1/2 degree
            # reverse direction
            logger.debug("set_direction_angle: reverse direction( nearly 180 degrees)")

            if start_edge.is_arc():
                start_lefthand = start_edge.n1_direction_lefthand()
                if self.is_arc():
                    myself_lefthand = self.n1_direction_lefthand()
                    if start_lefthand:
                        if not myself_lefthand:  # righthand => same side
                            left = self.radius() < start_edge.radius()
                        else:
                            left = False
                    else:
                        if myself_lefthand:  # lefthand => same side
                            left = self.radius() > start_edge.radius()
                        else:
                            left = True
                    if left:  # self is left
                        self.angle = np.pi * 2.0
                    else:
                        self.angle = 0.0
                elif self.is_line():
                    if start_lefthand:
                        self.angle = 0.0
                    else:
                        self.angle = np.pi * 2.0
            else:  # start is a line
                if self.is_arc():
                    if self.n1_direction_righthand():
                        self.angle = np.pi * 2.0
                    else:
                        self.angle = 0.0

        logger.debug("end set_direction_angle: angle is %s", self.angle)

    def n1_angle_outgoing(self):
        return self.element.get_alpha(self.n1)

    def n1_angle_ingoing(self):
        return normalise_angle(self.element.get_alpha(self.n1) + np.pi)

    def n2_angle_outgoing(self):
        return self.element.get_alpha(self.n2)

    def n2_angle_ingoing(self):
        return normalise_angle(self.element.get_alpha(self.n2) + np.pi)

    def n1_direction_lefthand(self):
        if not self.is_arc():
            return False
        alpha_in = self.n1_angle_ingoing()
        alpha_c = alpha_line(self.n1, self.element.center)
        angle = positive_angle(alpha_angle(alpha_in, alpha_c))
        return angle < np.pi

    def n1_direction_righthand(self):
        return not self.n1_direction_lefthand()

    def radius(self):
        if self.is_arc():
            return self.element.radius
        return 0.0

    def line_arc_close_together(self, info):
        assert(self.is_arc())
        assert(info.is_line())

        #if self.is_start_edge() or info.is_start_edge():
        #    return False

        dist_n2 = distance(self.element.center, info.n2)
        return np.isclose(self.element.radius, dist_n2, 0.005, 0.005)

    def myself_direction_lefthand(self, start_edge, nbr_edge, builder, ignore_start=False):
        logger.debug("start of myself_direction_lefthand")
        start_edge.log_edge("===> start")
        self.log_edge("---> self")
        nbr_edge.log_edge("---> nbr")

        myself_angle = self.angle
        other_angle = nbr_edge.angle

        if is_same_angle(0.0, myself_angle, atol=0.015):
            # 360 or 0 degrees => turn 180 degrees
            logger.debug("-- ATTENTION: myself %s turns nearly 180 degrees", self.classname())
            logger.debug("   the angle is %s", myself_angle)

            if start_edge.is_arc():
                if self.is_line() and not ignore_start:
                    if start_edge.line_arc_close_together(self):
                        logger.debug("START ARC and SELF LINE close")
                        arc_edge = start_edge.get_reverse_edge()  # reverse start edge
                        arc_edge.log_name("REVERSE")
                        arc_edge.set_direction_angle(start_edge)
                        left = not arc_edge.arc_line_direction_lefthand(start_edge, self, builder)
                        log_lefthand(left, self, arc_edge)
                        if left:  # self is left
                            myself_angle = 0.0
                        else:
                            myself_angle = np.pi * 2.0
                        left = myself_angle > other_angle

                        log_lefthand(left, self, nbr_edge)
                        logger.debug("#1a: end of myself_direction_lefthand: ==> %s", left)
                        return left
    
            elif self.is_arc():
                if not ignore_start:
                    if self.line_arc_close_together(start_edge):
                        logger.debug("START LINE and SELF ARC close")
                        line_edge = start_edge.get_reverse_edge()  # reverse start edge
                        line_edge.log_name("REVERSE")
                        line_edge.set_direction_angle(start_edge)
                        left = self.arc_line_direction_lefthand(start_edge, line_edge, builder)
                        log_lefthand(left, self, line_edge)
                        if left:  # self is left
                            myself_angle = 0.0
                        else:
                            myself_angle = np.pi * 2.0
                        left = myself_angle > other_angle

                        log_lefthand(left, self, nbr_edge)
                        logger.debug("#4: end of myself_direction_lefthand: ==> %s", left)
                        return left

        if is_same_angle(0.0, other_angle, atol=0.015):  # 1/2 degree
            # 360 or 0 degrees => turn 180 degrees
            logger.debug("-- ATTENTION: other %s turns nearly 180 degrees", nbr_edge.classname())
            logger.debug("   the angle is %s", other_angle)
            if start_edge.is_arc():
                if nbr_edge.is_line() and not ignore_start:
                    if start_edge.line_arc_close_together(nbr_edge):
                        logger.debug("START ARC and SELF LINE close")
                        arc_edge = start_edge.get_reverse_edge()  # reverse start edge
                        arc_edge.set_direction_angle(start_edge)
                        left = arc_edge.arc_line_direction_lefthand(start_edge, nbr_edge, builder)
                        log_lefthand(left, self, nbr_edge)
                        logger.debug("#5: end of myself_direction_lefthand: ==> %s", not left)
                        return not left

            elif nbr_edge.is_arc():
                if not ignore_start:
                    if nbr_edge.line_arc_close_together(start_edge):
                        logger.debug("START LINE and NEIGHBOR ARC close")
                        line_edge = start_edge.get_reverse_edge()  # reverse start edge
                        line_edge.set_direction_angle(start_edge)
                        left = nbr_edge.arc_line_direction_lefthand(start_edge, line_edge, builder)
                        logger.debug("#6: end of myself_direction_lefthand: ==> %s", left)
                        return left

        logger.debug("-- angles: myself = %s,  other = %s",
                     myself_angle, other_angle)
        if not is_same_angle(myself_angle, other_angle, atol=0.015):  # 1/2 degree
            logger.debug("-- angles are different")
            left = myself_angle > other_angle
            log_lefthand(left, self, nbr_edge)
            logger.debug("end of myself_direction_lefthand: ==> %s", left)
            return left

        logger.debug("nearly same direction")

        if self.is_line():
            if nbr_edge.is_line():
                return self.angle > nbr_edge.angle
            left = nbr_edge.arc_line_direction_lefthand(start_edge, self, builder)
            log_lefthand(left, nbr_edge, self)
            logger.debug("#1: end of myself_direction_lefthand: ==> %s", not left)
            return not left

        if nbr_edge.is_line():
            left = self.arc_line_direction_lefthand(start_edge, nbr_edge, builder)
            log_lefthand(left, self, nbr_edge)
            logger.debug("#2: end of myself_direction_lefthand: ==> %s", left)
            return left

        # two arcs
        left = self.myself_lefthand_side(nbr_edge)
        log_lefthand(left, self, nbr_edge)
        logger.debug("end of myself_direction_lefthand: ==> %s", left)
        return left

    def myself_lefthand_side(self, edge):
        # two edges with same direction
        if self.is_arc():
            if edge.is_line():
                return self.n1_direction_lefthand()
            
            edge_left = edge.n1_direction_lefthand()
            if self.n1_direction_lefthand():
                if not edge_left:
                    return True
                return self.radius() < edge.radius()
            else:
                if edge_left:
                    return False
                return self.radius() > edge.radius()
        else:
            if edge.is_arc():
                return edge.n1_direction_righthand()
        return True  # two lines

    def arc_line_direction_lefthand(self, start_edge, line_edge, builder):
        logger.debug("begin of arc_line_direction_lefthand")
        if not self.is_arc():
            logger.critical("FATAL: unexpected %s at position %s", self.classname(), self.n1)
            sys.exit(1)

        assert(self.is_arc())
        assert(line_edge.is_line())
        start_edge.log_edge(":::: start")
        self.log_edge(".... self")
        line_edge.log_edge(".... line")

        if not self.line_arc_close_together(line_edge):
            logger.debug("-- line and arc not close together")
            left = self.n1_direction_lefthand()  # ARC
            logger.debug("end of arc_line_direction_lefthand: arc lefthand = %s", left)
            return left

        logger.debug("-- ATTENTION: line look ahead")
        line_edge.startangle = start_edge.startangle
        # Line is start-edge now
        next_edge = builder.next_edge_lefthand_side(line_edge)  # next of line
        next_edge.set_direction_angle(start_edge)
        left = self.myself_direction_lefthand(start_edge,
                                              next_edge,
                                              builder,
                                              ignore_start=True)
        logger.debug("end of arc_line_direction_lefthand: arc lefthand = %s", left)
        return left

    def get_alpha_points(self, prev_n):
        n1 = self.n1
        n2 = self.n2
        a = 0.0
        if self.is_arc():
            angle = self.element.get_angle_of_arc()
            if angle > np.pi * 0.75:
                n = self.element.center_of_connection()
                a = normalise_angle(alpha_points(prev_n,
                                                 n1,
                                                 n))
                n1 = n

        a += normalise_angle(alpha_points(prev_n,
                                          n1,
                                          n2))
        return a, n1

class AreaBuilder(object):
    def __init__(self,
                 geom=None,
                 nolog=True,
                 rtol=1e-04,
                 atol=1e-04,
                 ndec=6):
        assert(geom is not None)
        self.rtol = rtol
        self.atol = atol
        self.ndec = ndec
        self.geom = geom
        self.area_list = []
        self.errors = 0
        self.journal = getJournal()
        self.nolog = nolog
        self.num_edges = self.geom.number_of_edges()

    def __str__(self):
        return "rtol: {}\n".format(self.rtol) + \
               "atol: {}\n".format(self.atol)

    def all_neighbors(self, n):
        for n in self.geom.g.neighbors(n):
            yield n

    def set_edge_attributes(self):
        self.geom.set_edge_attributes()

    def append_new_area(self, area_list, elements):
        a = Area(elements, self.geom.center, 0.0)
        a.set_type(TYPE_AIR)  # air
        area_list.append(a)
        return a

    def create_list_of_areas(self, main=False):
        logger.debug("Begin of create_list_of_areas")
        assert(len(self.area_list) == 0)
        self.errors = 0

        def append(area_list, a):
            for area in area_list:
                if area.is_identical(a):
                    return
            area_list.append(a)

        timer = Timer(start_it=True)
        if self.nolog:
            disable_logging()
        self.set_edge_attributes()
        
        for n in self.geom.nodes():
            finished = False
            while not finished:
                finished = True
                nbrs = [nbr for nbr in self.geom.g.neighbors(n)]
                for next_n in nbrs:
                    result = self.get_new_area(n, next_n)
                    if result['ok']:
                        a = self.append_new_area(self.area_list, result['area'])
                        logger.debug("Area %s found", a.identifier())
        if self.nolog:
            enable_logging()
        t = timer.stop("{} areas created in %0.4f seconds".format(len(self.area_list)))
        if main:
            self.journal.put('time_build_areas', t)
        if self.errors > 0:
            logger.warning("WARNING: %s errors while creating %s areas",
                           self.errors, len(self.area_list))
        self.journal.put_areas(len(self.area_list))
        self.journal.put_area_errors(self.errors)
        logger.debug("End of create_list_of_areas: %s areas created",
                     len(self.area_list))

    def get_neighbors(self, info):
        nbrs = [n for n in self.all_neighbors(info.n2)
                if not nodes_are_equal(n, info.n1)]
        if len(nbrs) == 0:
            logger.error("FATAL ERROR: no neighbors for node %s available ???",
                         info.n2)
            return []

        neighbors = []
        for n in nbrs:
            neighbors.append(self.get_edge_info(info.n2, n))
        return neighbors
    
    def get_edge_info(self, n1, n2):
        edge_data = self.geom.g.get_edge_data(n1, n2)
        if not edge_data:
            raise ValueError("Fatal: no edge-data found from {} to {}"
                             .format(n1, n2))

        return EdgeInfo(edge_data, n1, n2)

    def get_new_area(self, start_n1, start_n2):
        info_curr = self.get_edge_info(start_n1, start_n2)
        result = {'area': None,
                  'elements': 1,
                  'msg': "<undefined>",
                  'reverse': False,
                  'ok': False}
 
        if info_curr.is_tracked():
            result['msg'] = ("<== area already tracked (%s) ***",
                             info_curr.x)
            return result

        area = [info_curr.element]
        result['area'] = area

        logger.debug('begin of get_new_area')

        if info_curr.is_circle():
            result['msg'] = "area is a circle !!"
            logger.debug("<== %s", result['msg'])
            info_curr.set_all_tracked()
            result['ok'] = True
            result['circle'] = True
            return result

        logger.debug("***** EDGE %s *****", 1)
        info_curr.set_start_angle()
        info_curr.log_name("START")
        info_next = self.next_edge_lefthand_side(info_curr)
        if not info_next:
            result['msg'] = ("dead end ({}, {})"
                             .format(info_curr.n1, info_curr.n2))
            logger.debug("<== %s", result['msg'])
            self.errors += 1
            return result

        prev_node = info_curr.n1
        next_n1 = info_next.n1
        next_n2 = info_next.n2

        alpha, prev_node = info_next.get_alpha_points(prev_node)

        logger.debug("--> alpha %s <--", alpha)
        c = 1
        while not (nodes_are_equal(next_n1, start_n1) and
                   nodes_are_equal(next_n2, start_n2)):
            c += 1
            if c > self.num_edges * 2:
                logger.error("FATAL: *** over %s elements in area ? ***",
                             self.num_edges)
                sys.exit(1)

            area.append(info_next.element)

            if info_next.is_tracked():
                result['msg'] = ("FATAL: area already tracked ({}) ***"
                                 .format(info_next.x))
                logger.debug("<== %s", result['msg']) # ERROR
                logger.debug("++ %s from %s to %s",
                             info_next.classname(), info_next.n1, info_next.n2)
                result['elements'] = c
                self.errors += 1
                return result

            info_next.set_tracked()

            info_curr = info_next
            logger.debug("***** EDGE %s *****", c)
            info_next.set_start_angle()
            info_curr.log_name("START")
            info_next = self.next_edge_lefthand_side(info_curr)
            if not info_next:
                result['msg'] = ("<== dead end ({},{})"
                                 .format(info_curr.n1, info_curr.n2))
                logger.debug("<== %s", result['msg'])
                result['elements'] = c
                self.errors += 1
                return result

            next_n1 = info_next.n1
            next_n2 = info_next.n2

            a, prev_node = info_next.get_alpha_points(prev_node)
            alpha += a
            logger.debug("--> alpha %s (+ %s) <--", alpha, a)

        logger.debug("  END OF get_new_area")

        if less(alpha, 0.0):
            result['msg'] = ("turn left expected, but it turned right ({})"
                             .format(alpha))
            logger.debug("<== %s", result['msg'])
            result['elements'] = c
            result['reverse'] = True
            # it's not an error
            return result

        result['msg'] = "area found !!"
        logger.debug("<== %s", result['msg'])
        result['elements'] = c
        result['ok'] = True
        return result

    def next_edge_lefthand_side(self, start_edge):  # current
        logger.debug("begin of next_edge_lefthand_side")
        logger.debug("-- start is %s from %s to %s",
                     start_edge.classname(),
                     start_edge.n1,
                     start_edge.n2)

        nbrs = self.get_neighbors(start_edge)
        if len(nbrs) == 0:
            logger.debug("WARNING: no neighbors available ???")
            return None  # unexpected end => appendix

        if len(nbrs) == 1:
            logger.debug("end of next_edge_lefthand_side: one neighbor")
            return nbrs[0]

        logger.debug("-- NODE %s has %s neighbors", start_edge.n2, len(nbrs))
        logger.debug("-- startangle is %s", start_edge.startangle)

        all_nbrs = []
        logger.debug("candidates are:") 
        for i, nbr in enumerate(nbrs):
            nbr.set_direction_angle(start_edge)
            nbr.log_name("EDG-{}".format(i))
            nbr.log_edge("++++ candidate")
            all_nbrs.append((nbr.angle, i, nbr))
        logger.debug("--")
        all_nbrs.sort()

        a1, i1, nbr1 = all_nbrs[0]  # lefthandside
        nbr1.log_edge("==== first")
        for a2, i2, nbr2 in all_nbrs[1:]:
            nbr2.log_edge("==== next")
            if nbr2.myself_direction_lefthand(start_edge, nbr1, self):
                nbr1 = nbr2
                logger.debug("-- %s ist neuer lefthand neighbor", nbr1.name)
            else:
                logger.debug("-- %s bleibt lefthand neighbor", nbr1.name)
        nbr1.log_edge(">>>> LEFT")
        logger.debug("end of next_edge_lefthand_side")
        return nbr1

    def create_inner_corner_auxiliary_areas(self, startangle, endangle):
        logger.debug("begin of create_inner_corner_auxiliary_areas")
        if not self.geom.is_inner:
            logger.debug("end of create_inner_corner_auxiliary_areas: not inner")
            return False

        self.set_edge_attributes()

        start_nodes = [n for n in self.geom.angle_nodes(
            self.geom.center,
            startangle, 1e-3, 1e-3)]
        for n in start_nodes:
            nbrs = self.geom.get_neighbors(n)
            logger.debug("start node %s has %s neighbors", n, len(nbrs))
        logger.debug("corner nodes: %s", self.geom.start_corners)
        logger.debug("end nodes: %s", self.geom.end_corners)
        start_cp, start_exists = self.geom.get_start_airgap_corner()
        end_cp, end_exists = self.geom.get_end_airgap_corner()

        if start_exists and end_exists:
            logger.debug("end of create_inner_corner_auxiliary_areas: no aktion")
            return False

        airgap_line, airgap_el = self.get_inner_airgap_line()
        if not airgap_el:
            logger.debug("end of create_inner_corner_auxiliary_areas: no airgapline found")
            return False

        logger.debug("airgapline found !!")
        airgap_nodes = [n for n in airgap_line[1:]]
        del airgap_nodes[-1]

        created = False
        if not start_exists:
            cp = self.geom.start_corners[-1]
            logger.debug("Start Corner: %s -- %s", cp, start_cp)
            start_line = Line(Element(start=cp, end=start_cp),
                              color='red',
                              linestyle='dotted')

            start_cp = start_line.node2(self.ndec)
            for n in airgap_nodes:
                ag_line = Line(Element(start=start_cp, end=n))
                points = self.geom.get_intersection_points(airgap_el, ag_line, n)
                if not points:  # no intersection
                    d = distance(self.geom.center, n)
                    if np.isclose(d, self.geom.max_radius):
                        self.geom.add_arc(start_cp,
                                          n,
                                          self.geom.center,
                                          self.geom.max_radius,
                                          color='red',
                                          linestyle='dotted')
                    else:
                        self.geom.add_line(start_cp,
                                           n,
                                           color='red',
                                           linestyle='dotted')
                    created = True
                    start_node = self.geom.get_node(start_cp)
                    self.geom.add_edge(cp, start_node, start_line)
                    result = self.get_new_area(start_node, n)
                    if result['ok']:
                        self.append_new_area(self.geom.area_list,
                                             result['area'])
                    self.geom.set_start_corners(self.geom.center, 0.0)
                    break

        if not end_exists:
            cp = self.geom.end_corners[-1]
            logger.debug("End Corner: %s -- %s", cp, end_cp)
            end_line = Line(Element(start=cp, end=end_cp),
                            color='red',
                            linestyle='dotted')
            end_cp = end_line.node2(self.ndec)
            airgap_nodes.reverse()
            for n in airgap_nodes:
                ag_line = Line(Element(start=end_cp, end=n))
                points = self.geom.get_intersection_points(airgap_el, ag_line, n)
                if not points:  # no intersection
                    d = distance(self.geom.center, n)
                    if np.isclose(d, self.geom.max_radius):
                        self.geom.add_arc(n, end_cp,
                                          self.geom.center,
                                          self.geom.max_radius,
                                          color='red',
                                          linestyle='dotted')
                    else:
                        self.geom.add_line(end_cp, n,
                                           color='red',
                                           linestyle='dotted')
                    created = True
                    end_node = self.geom.get_node(end_cp)
                    self.geom.add_edge(cp, end_node, end_line)
                    result = self.get_new_area(n, end_node)
                    if result['ok']:
                        self.append_new_area(self.geom.area_list,
                                             result['area'])
                    self.geom.set_end_corners(self.geom.center, self.geom.alfa)
                    break

        logger.debug("end of create_inner_corner_auxiliary_areas")
        return created

    def get_airgap_line(self, n1, n2, end_node):
        logger.debug("get_airgap_line")

        self.set_edge_attributes()

        logger.debug("START EDGE IS: %s - %s", n1, n2)
        nodes = [n1, n2]
        info = self.get_edge_info(n1, n2)
        elements = [info.element]

        while not points_are_close(end_node, n2):
            info.set_start_angle()
            info = self.next_edge_lefthand_side(info)
            if not info:  # bad
                return [], []
            n2 = info.n2
            nodes.append(n2)
            elements.append(info.element)

        logger.debug("end of get_airgap_line: %s nodes", len(nodes))
        return nodes, elements

    def get_upper_border_line(self):
        start_node = self.geom.end_corners[-1]
        end_node = self.geom.start_corners[-1]
        next_node = self.get_right_neighbor(start_node)
        logger.debug("UPPER NODES start=%s, next=%s, end=%s", start_node, next_node, end_node)
        return self.get_airgap_line(start_node, next_node, end_node)

    def get_lower_border_line(self):
        start_node = self.geom.start_corners[0]
        end_node = self.geom.end_corners[0]
        next_node = self.get_left_neighbor(start_node)
        logger.debug("LOWER NODES start=%s, next=%s, end=%s", start_node, next_node, end_node)
        return self.get_airgap_line(start_node, next_node, end_node)

    def get_inner_airgap_line(self):
        logger.debug("get_inner_airgap_line()")
        assert(self.geom.is_inner)
        assert(self.geom.area_list)
        return self.get_upper_border_line()

    def get_inner_opposite_airgap_line(self):
        logger.debug("get_inner_opposite_airgap_line()")
        assert(self.geom.is_inner)
        assert(self.geom.area_list)
        return self.get_lower_border_line()

    def close_outer_winding_areas(self, color='red'):
        logger.debug("begin close_outer_winding_areas")

        airgap_line, airgap_el = self.get_outer_airgap_line()
        logger.debug("Outer Airgap with %s Nodes", len(airgap_line))

        if len(airgap_line) < 5:
            return False

        distlist = [distance(self.geom.center, n) for n in airgap_line]
        min_dist = min(distlist)
        max_dist = max(distlist)
        geom_height = self.geom.max_radius - self.geom.min_radius
        line_height = max_dist - min_dist
        if line_height < geom_height * 0.2:
            return 0

        logger.debug("NODES: %s TO %s", airgap_line[0], airgap_line[-1])
        n1 = None
        dist_n1 = 0.0

        e_prev = None
        n_prev = airgap_line[0]
        dist_prev = distance(self.geom.center, n_prev)
        alpha_prev = alpha_line(self.geom.center, n_prev)
        alpha_start = alpha_prev
        lines_created = 0
        dist_ag = self.geom.min_radius
        logger.debug("MIN RADIUS DIST: %s",dist_ag)
        for n in airgap_line[1:]:
            dist = distance(self.geom.center, n)
            alpha = alpha_line(self.geom.center, n)
            logger.debug("Node %s, dist=%s, alpha=%s", n, dist, alpha)
            if not n1:
                logger.debug("Previous, dist=%s, alpha=%s", dist_prev, alpha_prev)
                if greater(dist_prev, dist_ag, rtol=1e-3, atol=1e-3):
                    if greater(dist, dist_prev, rtol=1e-3, atol=1e-3) and \
                       less(alpha, alpha_prev, rtol=1e-4, atol=1e-5):
                        n1 = n_prev
                        dist_n1 = dist_prev
                        logger.debug("NODE1: %s  (%s)", n1, dist_n1)
            else:
                if less_equal(dist, dist_n1, rtol=1e-3, atol=1e-2):
                    logger.debug("NODE2: %s  (%s)", n, dist)
                    line = Line(Element(start=n1, end=n))
                    if e_prev.intersect_line(line):
                        logger.debug("___LINE NOT POSSIBLE___")
                    else:
                        self.geom.add_line(n1, n, color=color)
                        lines_created += 1
                    n1 = None
                    dist_n1 = 0.0
            if not n1:
                e_prev = self.geom.get_edge_element(n_prev, n)
            n_prev = n
            dist_prev = dist
            alpha_prev = alpha

        logger.debug("end close_outer_winding_areas: lines=%s", lines_created)
        return lines_created > 0

    def get_left_neighbor(self, start_node):
        nbrs = self.geom.get_neighbors(start_node)
        start_alpha = normalise_angle(alpha_line(start_node, self.geom.center))
        next_angle = 0.0
        next_node = None
        for n in nbrs:
            alpha = normalise_angle(alpha_line(start_node, n))
            angle = alpha_angle(start_alpha, alpha)
            if angle > next_angle:
                next_angle = angle
                next_node = n

        logger.debug("Left neighbor is %s", next_node)
        return next_node

    def get_right_neighbor(self, start_node):
        nbrs = self.geom.get_neighbors(start_node)
        start_alpha = normalise_angle(alpha_line(start_node, self.geom.center))
        next_angle = 99
        next_node = None
        for n in nbrs:
            alpha = normalise_angle(alpha_line(start_node, n))
            angle = alpha_angle(start_alpha, alpha)
            if angle < next_angle:
                next_angle = angle
                next_node = n

        logger.debug("Right neighbor is %s", next_node)
        return next_node

    def get_outer_airgap_line(self):
        logger.debug("get_outer_airgap_line()")
        assert(self.geom.is_outer)
        assert(self.geom.area_list)
        return self.get_lower_border_line()

    def get_outer_opposite_airgap_line(self):
        logger.debug("get_outer_opposite_airgap_line()")
        assert(self.geom.is_outer)
        assert(self.geom.area_list)
        return self.get_upper_border_line()

        # same procedure as inner airgap line
        return self.get_inner_airgap_line()

    def create_one_area_group(self, areas):
        logger.debug("begin of create_one_area_group")
        assert(len(self.area_list) == 0)

        self.set_edge_attributes()
        if not self.build_group_area(areas):
            logger.warning("Creation of an areagroup failed")
            return True
        return False

    def create_area_groups(self, list_of_areas):
        logger.debug("begin of create_area_groups")
        assert(len(self.area_list) == 0)

        area_id_list = [a.id for a in list_of_areas]
        timer = Timer(start_it=True)
        #   ---
        def delete_id(id):
            try:
                i = area_id_list.index(id)
            except ValueError:
                return True
            area_id_list[i] = 0
        #   ---

        group_list = {}
        for area in list_of_areas:
            if not area.id in area_id_list:
                continue

            areas = [area]
            delete_id(area.id)
            for a in list_of_areas:
                if area.id == a.id:
                    continue
                if not a.id in area_id_list:
                    continue
                if area.is_in_touch_with_area(self.geom, a):
                    areas.append(a)
                    delete_id(a.id)

            group_list[area.id] = areas

        group_keys = [int(x) for x in group_list.keys()]
        logger.debug("=== Area Group List ===")
        self.set_edge_attributes()
        errors = 0
        disable_logging()
        for k in group_keys:
            if not self.build_group_area(group_list[k]):
                logger.warning("Creation of an areagroup failed")
                errors += 1
        enable_logging()

        t = timer.stop("areagroups created in %0.4f seconds")
        logger.debug("end of create_area_groups: %s groups created",
                     len(self.area_list))
        return errors > 0

    def build_group_area(self, area_list):
        id_list = [a.id for a in area_list]
        logger.debug("Area Group: %s", id_list)

        max_x = 0
        area = None
        for a in area_list:
            if a.is_one_circle_element():
                continue
            if a.max_x > max_x:
                max_x = a.max_x
                area = a
        if area is None:
            return False

        x0, y0 = -9999.0, 0.0
        for x, y in area.list_of_nodes():
            if x > x0:
                x0 = x
                y0 = y

        node0 = (x0, y0)
        if not self.geom.g.has_node(node0):
            node0 = self.geom.find_the_node(node0)
            x0, y0 = node0

        nbrs = self.geom.get_neighbors(node0)
        logger.debug("Neighbors of %s", node0)

        alpha0 = alpha_line(node0, (x, -9999.0))
        angle0 = np.pi * 2.0
        for n in nbrs:
            logger.debug("** nbr: %s", n)
            alpha1 = alpha_line(node0, n)
            angle1 = alpha_angle(alpha1, alpha0)
            logger.debug("** nbr=%s,  alpha=%s,  angle=%s", n, alpha1, angle1)
            logger.debug("** if %s < %s then", angle1, angle0)
            if angle1 < angle0:
                node1 = n
                angle0 = angle1

        logger.debug("** Start with nodes %s ==> %s", node0, node1)
        rslt = self.get_new_area(node0, node1)
        if rslt['ok'] and rslt.get('circle', False):
            rslt['ok'] = False
            rslt['reverse'] = True

        logger.debug("** Result: ok=%s, reverse=%s", rslt['ok'], rslt['reverse'])
        if not rslt['ok'] and rslt['reverse']:
            # new areagroup
            a = self.append_new_area(self.area_list, rslt['area'])
            a.areas_of_group = area_list
            return True
        return False



#!/usr/bin/env python
#
import femagtools.poc
import pathlib

sinPars = dict(
    shape_current="sin",
    phi_voltage_winding=[0.0, 120.0, 240.0],
    skew_angle=30.0,
    key_winding=[1, 2, 3],
    num_skew_steps=3)

funPars = dict(
    shape_current="sin",
    phi_voltage_winding=[-80.0, 40.0, 160.0],
    key_winding=['1', '2', '3'],
    func_current=[0.0,
                   11.25,
                   11.25,
                   33.75,
                   33.75,
                   56.25,
                   56.25,
                   78.75,
                   78.75,
                   101.25,
                   101.25,
                   123.75,
                   123.75,
                   146.25,
                   146.25,
                   168.25,
                   168.25,
                   191.25,
                   191.25,
                   213.75,
                   213.75,
                   236.25,
                   236.25,
                   258.75,
                   258.75,
                   281.25,
                   281.25,
                   303.75,
                   303.75,
                   326.25,
                   326.25,
                   348.75,
                   348.75,
                   360.0]
    ,func_phi= [0.0,
                0.0,
                0.4,
                0.4,
                1.1,
                1.1,
                1.66,
                1.66,
                2.0,
                2.0,
                1.66,
                1.66,
                1.1,
                1.1,
                0.4,
                0.4,
                0.0,
                0.0,
                -0.4,
                -0.4,
                -1.1,
                -1.1,
                -1.66,
                -1.66,
                -2.0,
                -2.0,
                -1.66,
                -1.66,
                -1.1,
                -1.1,
                -0.4,
                -0.4,
                0.0,
                0.0])

funcPars = dict(
        func_current=[0.0,0.0,1.0,1.0,0.0,0.0,-1.0,-1.0,0.0,0.0]
        ,func_phi=[0.0,30.0,30.0,150.0,150.0,210.0,210.0,330.0,330.0,360.0]
        ,harmonic_id=[1,3,5,7,9,11,13,15,17,19])


def test_default():
    poles = 4
    pole_pitch = 2*360/poles
    poc = femagtools.poc.Poc(pole_pitch)
    assert poc.filename() == f'sin_{poles}p.poc'
    expected = ['3','1','2','3','0.0','120.0','240.0',f'{pole_pitch}','sin','0.0','0','']
    assert poc.content() == expected


def test_write_poc(tmpdir):
    poc = femagtools.poc.Poc(360, sinPars)
    filename = tmpdir / 'sin.poc'

    poc.write(filename)
    expected = '3\n1\n2\n3\n0.0\n120.0\n240.0\n360\nsin\n30.0\n3\n'
    assert filename.read() == expected


def test_read_poc():
    datadir = pathlib.Path(__file__).resolve().parent.joinpath('data')
    poc = femagtools.poc.Poc(str(datadir / 'test.poc'))
    expected = sinPars
    expected['pole_pitch'] = 360.0
    expected['pocType'] = 'Function'
    assert poc.getProps() == expected

def test_read_func_poc():
    datadir = pathlib.Path(__file__).resolve().parent.joinpath('data')
    poc = femagtools.poc.Poc(str(datadir / 'test_func.poc'))
    assert poc.getProps()['pole_pitch'] == 18.0
    assert poc.getProps()['pocType'] == 'fun'

#    def test_read_poc_with_fun( self ):
#        poc=self._createPoc('2p_sin.poc')
#        self.assertEqual(poc.getProps(), funPars)

def test_rec(tmpdir):
    key_winding=['1','2', '3']
    phi_voltage_winding=[30, 150, 270]
    p = 2
    poc = femagtools.poc.Poc(360/p,
                             dict(key_winding=key_winding,
                                  phi_voltage_winding=phi_voltage_winding,
                                  shape_current='rec'))
    filename = tmpdir / 'rec.poc'
    poc.write(filename)
    expected = '3\n1\n2\n3\n30\n150\n270\n180.0\nrec\n0.0\n0\n'
    assert filename.read() == expected

def test_har(tmpdir):
    key_winding=['1','2', '3']
    phi_voltage_winding=[30, 150, 270]
    p = 2
    poc = femagtools.poc.Poc(360/p,
                             dict(key_winding=key_winding,
                                  phi_voltage_winding=phi_voltage_winding,
                                  func_current=[1, 0, 0.3, 0, 0.1],
                                  func_phi=[0, 0, 0, 0, 0],
                                  pocType='har'))
    filename = tmpdir / 'har.poc'
    poc.write(filename)
    expected = '3\n1\n2\n3\n30\n150\n270\n180.0\nhar\n5\n1, 0\n0, 0\n0.3, 0\n0, 0\n0.1, 0\n0.0\n0\n'
    assert filename.read() == expected

def test_hsp(tmpdir):
    poc = femagtools.poc.HspPoc(harm=[1,21],
                                amp=[1, 0.01],
                                phi=[0, 0])
    poc.pole_pitch = 360/2
    #poc = femagtools.poc.Poc(360/p,
    #                         dict(harmonic_id=[1, 21],
    #                              func_current=[1, 0.01],
    #                              func_phi=[0, 0],
    #                              pocType='hsp'))
    filename = tmpdir / 'hsp.poc'
    poc.write(filename)

    expected = '3\n1\n2\n3\n0.0\n120.0\n240.0\n180.0\nhsp\n2\n1, 1, 0\n21, 0.01, 0\n0.0\n0\n'
    assert filename.read() == expected

def test_fun(tmpdir):
    p = 2
    key_winding=['1','2', '3']
    phi_voltage_winding=[30, 150, 270]
    poc = femagtools.poc.Poc(360/p,
                             dict(key_winding=key_winding,
                                  phi_voltage_winding=phi_voltage_winding,
                                  func_current=[0, 1, -1, 0],
                                  func_phi=[0, 45, 135, 180],
                                  pocType='fun'))
    filename = tmpdir / 'fun.poc'
    poc.write(filename)
    expected = '3\n1\n2\n3\n30\n150\n270\n180.0\nfun\n4\n0, 0\n45, 1\n135, -1\n180, 0\n0, 0\n45, 1\n135, -1\n180, 0\n0, 0\n45, 1\n135, -1\n180, 0\n0.0\n0\n'



# -*- coding: utf-8 -*-
"""
    femagtools.conductor
    ~~~~~~~~~~~~~~~~~~~~

    Creating and managing coonductor material


"""


class Conductor:
    def __init__(self, clist=[]):
        """initialize this object from a list of conductor objects"""
        self.conductors = {}
        for c in clist:
            if 'id' in c:
                self.conductors[str(c['id'])] = c
            elif 'name' in c:
                self.conductors[c['name']] = c

    def find(self, id):
        """find conductor by id or name"""
        try:
            return self.conductors[id]
        except ValueError:
            pass  # not found
        except KeyError:
            pass
        return self.find_by_name(id)

    def find_by_name(self, name):
        """find conductor by name"""
        for k in self.conductors.keys():
            if self.conductors[k]['name'] == name:
                return self.conductors[k]
            # not found
        return None

    def __str__(self):
        return str([self.conductors[c]
                    for c in self.conductors])



"""Python API for FEMAG

"""
__title__ = 'femagtools'
__version__ = '1.8.19'
__author__ = 'Ronald Tanner'
__license__ = 'BSD'
__copyright__ = 'Copyright 2023-2025 Gamma Technology'

from .asm import read
from .bch import Reader
from .model import MachineModel
from .fsl import Builder
from .magnet import Magnet
from .femag import Femag, ZmqFemag
from .conductor import Conductor


def read_bchfile(filename):
    """Read BCH/BATCH results from file *filename*."""
    import io
    bchresults = Reader()
    with io.open(filename, encoding='latin1', errors='ignore') as f:
        bchresults.read(f.readlines())
    return bchresults


def create_fsl(machine,
               operatingconditions={},
               magnetmat=[],
               condMat=[],
               templatedirs=[]):
    """create FSL command list from model and operating parameters

    Args:
        machine: dict with parameters (model)
        operatingConditions: dict with parameters (simulation)
        magnetmat: list fo dict with parameters (magnet material)
        templatedir: additional template directory
"""
    model = MachineModel(machine)

    builder = Builder(templatedirs)
    magnets = []
    if magnetmat:
        magnets = Magnet(magnetmat)
    if condMat:
        if not isinstance(condMat, Conductor):
            condMat = Conductor(condMat)

    if operatingconditions:
        return builder.create(model, operatingconditions, magnets)
    return builder.create_model(model,
                                condMat=condMat,
                                magnets=magnets) + ['save_model("cont")']



#!/usr/bin/env python
#
import pytest
import femagtools.grid
import femagtools.parstudy
import numpy as np
import functools


def test_create_parameter_range():
    x = [(1, 2, 3), (4, 5), (6, 7)]

    r = femagtools.grid.create_parameter_range(x)

    assert r.tolist() == [[1, 4, 6],
                          [2, 4, 6],
                          [3, 4, 6],
                          [1, 5, 6],
                          [2, 5, 6],
                          [3, 5, 6],
                          [1, 4, 7],
                          [2, 4, 7],
                          [3, 4, 7],
                          [1, 5, 7],
                          [2, 5, 7],
                          [3, 5, 7]]


@pytest.fixture
def decision_vars():
    return [
        {"steps": 3, "bounds": [-50, 0],
         "name": "angl_i_up", "label":"Beta"},
        {"steps": 3, "bounds": [100, 200],
         "name": "current", "label":"Current/A"}
    ]


def test_baskets():
    x = list(range(5))*5
    baskets = femagtools.parstudy.baskets(x, 5)

    for i, p in enumerate(baskets):
        assert p == [0, 1, 2, 3, 4]
    assert i == 4


def test_report(decision_vars):
    objective_vars = [
        {"name": "dqPar.torque[-1]",
         "label": "Load Torque/Nm"},
        {"name": "torque[-1].ripple",
         "label": "Torque Ripple/Nm"},
        {"name": "machine.plfe[-1]",
         "label": "Iron Losses/W"}
    ]

    shape = [len(objective_vars)] + [d['steps'] for d in decision_vars]
    a = [0]*len(objective_vars)*functools.reduce(
        (lambda x, y: x * y), [d['steps'] for d in decision_vars])
    objectives = np.reshape(a, shape)

    expected = [[d['label'] for d in decision_vars] +
                [o['label'] for o in objective_vars] + ['Directory'],
                [d['name'] for d in decision_vars] +
                [o['name'] for o in objective_vars],
                [-50.0, 100.0, 0.0, 0.0, 0.0, 0],
                [-25.0, 100.0, 0.0, 0.0, 0.0, 1],
                [-0.0,  100.0, 0.0, 0.0, 0.0, 2],
                [-50.0, 150.0, 0.0, 0.0, 0.0, 3],
                [-25.0, 150.0, 0.0, 0.0, 0.0, 4],
                [-0.0,  150.0, 0.0, 0.0, 0.0, 5],
                [-50.0, 200.0, 0.0, 0.0, 0.0, 6],
                [-25.0, 200.0, 0.0, 0.0, 0.0, 7],
                [-0.0, 200.0, 0.0, 0.0, 0.0, 8]]

    domain = [list(np.linspace(d['bounds'][0], d['bounds'][1], d['steps']))
              for d in decision_vars]
    par_range = femagtools.grid.create_parameter_range(domain)
    assert expected == femagtools.parstudy.get_report(decision_vars,
                                                      objective_vars, objectives,
                                                      par_range)


def test_sobol(decision_vars):
    parvar = femagtools.parstudy.Sobol('.')
    N = 4
    n, d, r = parvar._get_names_and_range(decision_vars, N)
    assert [d['name'] for d in decision_vars] == n
    assert r.shape == (N, len(decision_vars))


def test_lhs(decision_vars):
    parvar = femagtools.parstudy.LatinHypercube('.')
    N = 4
    n, d, r = parvar._get_names_and_range(decision_vars, N)
    assert [d['name'] for d in decision_vars] == n
    assert r.shape == (N, len(decision_vars))
